#
#   Imandra Inc.
#
#   speclogician/presentation/preview/fixtures/instances_list.py
#

from __future__ import annotations

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.instances_list import (
    InstancesListPM,
    InstancesListRowPM,
)
from speclogician.presentation.models.state_instance_summary import StateInstanceSummaryPM
from speclogician.utils import IMLValidity
from speclogician.modeling.domain import BaseStatus


# -----------------------------------------------------------------------------
# helpers
# -----------------------------------------------------------------------------

def _summary(
    *,
    created_at: str,
    num_changes: int,
    base_status: BaseStatus = BaseStatus.UNKNOWN,
    base_has_state: bool = False,
    base_has_action: bool = False,
    num_preds_total: int = 0,
    num_preds_matched: int = 0,
    num_trans_total: int = 0,
    num_trans_valid_logic: int = 0,
    num_sc_total: int = 0,
    num_sc_missing: int = 0,
    num_sc_inconsistent: int = 0,
    num_sc_conflicted: int = 0,
    num_test_traces_total: int = 0,
    num_test_traces_logic_good: int = 0,
    num_log_traces_total: int = 0,
    num_log_traces_logic_good: int = 0,
) -> StateInstanceSummaryPM:
    return StateInstanceSummaryPM(
        created_at=created_at,
        num_changes=num_changes,
        base_status=base_status,
        base_has_state=base_has_state,
        base_has_action=base_has_action,
        num_preds_total=num_preds_total,
        num_preds_matched=num_preds_matched,
        num_trans_total=num_trans_total,
        num_trans_valid_logic=num_trans_valid_logic,
        num_sc_total=num_sc_total,
        num_sc_missing=num_sc_missing,
        num_sc_inconsistent=num_sc_inconsistent,
        num_sc_conflicted=num_sc_conflicted,
        num_test_traces_total=num_test_traces_total,
        num_test_traces_logic_good=num_test_traces_logic_good,
        num_log_traces_total=num_log_traces_total,
        num_log_traces_logic_good=num_log_traces_logic_good,
    )


# -----------------------------------------------------------------------------
# fixtures
# -----------------------------------------------------------------------------

def minimal_instances_list_pm() -> InstancesListPM:
    return InstancesListPM(rows=[])


def typical_instances_list_pm() -> InstancesListPM:
    return InstancesListPM(
        rows=[
            InstancesListRowPM(
                state_label="latest",
                changes="DomainModelBaseEdit, PredicateAdd",
                summary=_summary(
                    created_at="2026-01-09T10:12:00Z",
                    num_changes=2,
                    base_status=BaseStatus.VALID,
                    base_has_state=True,
                    base_has_action=True,
                    num_preds_total=12,
                    num_preds_matched=7,
                    num_trans_total=4,
                    num_trans_valid_logic=3,
                    num_sc_total=9,
                    num_sc_missing=1,
                    num_test_traces_total=8,
                    num_test_traces_logic_good=6,
                    num_log_traces_total=4,
                    num_log_traces_logic_good=4,
                ),
            ),
            InstancesListRowPM(
                state_label="1",
                changes="PredicateEdit",
                summary=_summary(
                    created_at="2026-01-09T09:55:10Z",
                    num_changes=1,
                    base_status=BaseStatus.VALID,
                    base_has_state=True,
                    base_has_action=True,
                    num_preds_total=11,
                    num_preds_matched=6,
                    num_trans_total=4,
                    num_trans_valid_logic=3,
                    num_sc_total=8,
                    num_sc_missing=0,
                    num_test_traces_total=7,
                    num_test_traces_logic_good=6,
                    num_log_traces_total=3,
                    num_log_traces_logic_good=3,
                ),
            ),
        ]
    )


def edge_instances_list_pm() -> InstancesListPM:
    return InstancesListPM(
        rows=[
            InstancesListRowPM(
                state_label="latest",
                changes="",
                summary=_summary(
                    created_at="2026-01-09T10:12:00.123456Z",
                    num_changes=0,
                    base_status=BaseStatus.UNKNOWN,
                    base_has_state=False,
                    base_has_action=False,
                    num_preds_total=0,
                    num_preds_matched=0,
                    num_trans_total=0,
                    num_trans_valid_logic=0,
                    num_sc_total=0,
                    num_sc_missing=0,
                    num_test_traces_total=0,
                    num_test_traces_logic_good=0,
                    num_log_traces_total=0,
                    num_log_traces_logic_good=0,
                ),
            )
        ]
    )

def make_instances_list_pm(ctx: RenderCtx, variant: str = "typical") -> InstancesListPM:
    v = (variant or "typical").strip().lower()

    if v == "minimal":
        return minimal_instances_list_pm()

    if v == "edge":
        return edge_instances_list_pm()

    return typical_instances_list_pm()


def render(pm: InstancesListPM, ctx : RenderCtx):
    from speclogician.presentation.renderers.instances_list import render_instances_list
    return render_instances_list(pm, ctx=ctx)


#
#   Imandra Inc.
#
#   speclogician/presentation/builders/state_instance_summary.py
#

from __future__ import annotations

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.state_instance_summary import StateInstanceSummaryPM
from speclogician.state.inst import StateInstance


def _g(obj: object, name: str, default):
    return getattr(obj, name, default)


def build_state_instance_summary_pm(si: StateInstance, ctx: RenderCtx) -> StateInstanceSummaryPM:
    dm = si.spec.domain_model
    spec = si.spec
    art = si.art_container

    return StateInstanceSummaryPM(
        created_at=si.created_at.isoformat() if si.created_at else None,
        num_changes=len(si.changes or []),

        # ---- Domain model (base) ----
        base_status=str(_g(dm, "base_status", None)),
        base_has_state=_g(dm, "base_has_state", False),
        base_has_action=_g(dm, "base_has_action", False),

        # ---- State predicates ----
        num_state_preds_total=_g(dm, "num_state_preds_total", 0),
        num_state_preds_matched=_g(dm, "num_state_preds_matched", 0),
        num_state_preds_valid_logic=_g(dm, "num_state_preds_valid_logic", 0),
        num_state_preds_errored=_g(dm, "num_state_preds_errored", 0),

        # ---- Action predicates ----
        num_action_preds_total=_g(dm, "num_action_preds_total", 0),
        num_action_preds_matched=_g(dm, "num_action_preds_matched", 0),
        num_action_preds_valid_logic=_g(dm, "num_action_preds_valid_logic", 0),
        num_action_preds_errored=_g(dm, "num_action_preds_errored", 0),

        # ---- Predicates totals ----
        num_preds_total=_g(dm, "num_preds_total", 0),
        num_preds_matched=_g(dm, "num_preds_matched", 0),
        num_preds_valid_logic=_g(dm, "num_preds_valid_logic", 0),
        num_preds_errored=_g(dm, "num_preds_errored", 0),

        # ---- Transitions ----
        num_trans_total=_g(dm, "num_trans_total", 0),
        num_trans_matched=_g(dm, "num_trans_matched", 0),
        num_trans_valid_logic=_g(dm, "num_trans_valid_logic", 0),
        num_trans_errored=_g(dm, "num_trans_errored", 0),

        # ---- Scenarios ----
        num_sc_total=_g(spec, "num_sc_total", 0),
        num_sc_missing=_g(spec, "num_sc_missing", 0),
        num_sc_matched=_g(spec, "num_sc_matched", 0),
        num_sc_inconsistent=_g(spec, "num_sc_inconsistent", 0),

        # ---- Conflicts (spec-level) ----
        num_sc_conflicted=_g(spec, "num_sc_conflicted", 0),
        num_sc_overlap=_g(spec, "num_sc_overlap", 0),
        num_sc_consumed=_g(spec, "num_sc_consumed", 0),

        # ---- ArtifactContainer: test traces ----
        num_test_traces_total=_g(art, "num_test_traces_total", 0),
        num_test_traces_logic_good=_g(art, "num_test_traces_logic_good", 0),
        num_test_traces_matched=_g(art, "num_test_traces_matched", 0),

        # ---- ArtifactContainer: log traces ----
        num_log_traces_total=_g(art, "num_log_traces_total", 0),
        num_log_traces_logic_good=_g(art, "num_log_traces_logic_good", 0),
        num_log_traces_matched=_g(art, "num_log_traces_matched", 0),

        # ---- ArtifactContainer: src code ----
        num_src_code_arts_total=_g(art, "num_src_code_arts_total", 0),
        num_src_code_arts_matched=_g(art, "num_src_code_arts_matched", 0),

        # ---- ArtifactContainer: docs ----
        num_doc_arts_total=_g(art, "num_doc_arts_total", 0),
        num_doc_arts_matched=_g(art, "num_doc_arts_matched", 0),
    )


#
#   Imandra Inc.
#
#   speclogician/state/recommender.py
#

from __future__ import annotations

from typing import Iterable, TypeAlias, Callable, Optional

from pydantic import BaseModel

from .diff import ComparisonOutcome, ValueDiff, StateDiff
from .inst import StateInstance
from .recommendation import Recommendation

class ChangeFailure(BaseModel):
    """
    Lightweight failure context for recommendation generation.
    This replaces dependency on ProcessChangeResult.
    """

    change: str                 # e.g. "AddPredicate", "EditScenario"
    stage: str                  # e.g. "apply_change", "analyze_change"
    error_type: str             # e.g. "TypeError"
    error: str                  # str(exception)

    # Optional extras if you want later
    message: Optional[str] = None


Rule: TypeAlias = Callable[[StateInstance, StateInstance, Optional[ChangeFailure]], Iterable[Recommendation]]


class Recommender:
    """
    Turn (old_state, new_state) into human-readable next steps.

    Typical use:
      recs = Recommender().recommend(old_inst, new_inst)
      # feed "\n".join(r.text for r in recs) into your LLM CLI loop
    """

    def __init__(self, rules: list[Rule] | None = None) -> None:
        self._rules: list[Rule] = rules or [
            _rule_error_triage,
            _rule_state_diff_summary,
            _rule_missing_components,
            _rule_inconsistent_scenarios,
            _rule_conflicting_scenarios,
            _rule_unmatched_traces,
            _rule_invalid_iml_components,
            _rule_traceability_nudge,
        ]

    def recommend(
        self,
        old_inst: StateInstance,
        new_inst: StateInstance,
        *,
        failure: ChangeFailure | None = None,
    ) -> list[Recommendation]:
        out: list[Recommendation] = []

        for rule in self._rules:
            try:
                out.extend(list(rule(old_inst, new_inst, failure)))
            except Exception as e:
                # recommender should never take down the CLI
                out.append(
                    Recommendation(
                        text=f"Recommender rule failed ({getattr(rule, '__name__', type(rule).__name__)}): {e}",
                        kind="warning",
                        priority=5,
                    )
                )

        # Deduplicate by `text` while preserving order.
        # If the same text appears multiple times, keep the one with the lowest priority number.
        chosen: dict[str, Recommendation] = {}
        order: list[str] = []

        for r in out:
            if r.text not in chosen:
                chosen[r.text] = r
                order.append(r.text)
            else:
                prev = chosen[r.text]
                if r.priority < prev.priority:
                    chosen[r.text] = r

        return [chosen[t] for t in order]


# ----------------------------
# Helpers
# ----------------------------

def _ok(failure: ChangeFailure | None) -> bool:
    """
    Mirror old `res["ok"]` meaning:
    - failure is None => ok
    - failure provided => failed
    """
    return failure is None


def _outcome(v: ValueDiff[object]) -> ComparisonOutcome:
    """
    Use the comp_func if present, otherwise fallback to raw equality.
    """
    if v.comp_func is not None:  # type: ignore[truthy-function]
        return v.comp_func(v.before, v.after)  # type: ignore[misc]
    return ComparisonOutcome.NO_CHANGE if v.before == v.after else ComparisonOutcome.UNKNOWN


def _rec(
    text: str,
    *,
    kind: str = "next",
    priority: int = 50,
) -> Recommendation:
    return Recommendation(text=text, kind=kind, priority=priority)  # type: ignore[arg-type]


# ----------------------------
# Rules (keep them small + boring)
# ----------------------------

def _rule_error_triage(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if _ok(failure):
        return []

    # If you have the actual StateChange object at the call site,
    # you can set failure.change = ch.__class__.__name__.
    return [
        Recommendation(
            kind="warning",
            priority=0,
            text=(
                f"The last change failed during **{failure.stage}** for `{failure.change}`.\n"
                f"Error: `{failure.error_type}: {failure.error}`\n"
                "Recommendation: fix this error first (no later changes will be reliable)."
            ),
        )
    ]


def _rule_missing_components(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    spec = new_inst.spec
    missing = [
        sc
        for sc in spec.scenarios
        if getattr(sc, "preds_missing", []) or getattr(sc, "trans_missing", [])
    ]
    if not missing:
        return []

    examples = missing[:3]
    bits: list[str] = []
    for sc in examples:
        pm = getattr(sc, "preds_missing", [])
        tm = getattr(sc, "trans_missing", [])
        bits.append(f"- `{sc.name}` missing preds={pm or '—'} trans={tm or '—'}")

    return [
        Recommendation(
            text=(
                "Some scenarios reference predicates/transitions that do not exist yet.\n"
                + "\n".join(bits)
                + ("\n(…more)" if len(missing) > len(examples) else "")
                + "\nRecommendation: add the missing predicates/transitions or edit the scenarios to match the domain model."
            )
        )
    ]


def _rule_inconsistent_scenarios(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    spec = new_inst.spec
    bad = [sc for sc in spec.scenarios if getattr(sc, "all_preds_consistent", True) is False]
    if not bad:
        return []

    names = ", ".join(f"`{sc.name}`" for sc in bad[:5])
    tail = " …" if len(bad) > 5 else ""
    return [
        Recommendation(
            text=(
                f"Some scenarios appear **inconsistent** (unsat) given the current predicates: {names}{tail}.\n"
                "Recommendation: relax contradictory predicates, or split the scenario into smaller cases."
            )
        )
    ]


def _rule_conflicting_scenarios(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    spec = new_inst.spec
    conflicted = [sc for sc in spec.scenarios if getattr(sc, "conflicts", [])]
    if not conflicted:
        return []

    names = ", ".join(f"`{sc.name}`" for sc in conflicted[:5])
    tail = " …" if len(conflicted) > 5 else ""
    return [
        Recommendation(
            text=(
                f"Some scenarios have **overlaps/conflicts** with others: {names}{tail}.\n"
                "Recommendation: add distinguishing predicates (boundary conditions) so scenarios become disjoint, "
                "or explicitly document intended priority/ordering."
            )
        )
    ]


def _rule_unmatched_traces(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    cont = new_inst.art_container
    amap = new_inst.art_map

    if amap is None:
        return []

    def _is_trace_matched(trace) -> bool:
        tid = getattr(trace, "art_id", None)
        if not tid:
            return False
        return bool(amap.get_comp_for_artifact(tid))

    all_traces = list(cont.test_traces) + list(cont.log_traces)
    if not all_traces:
        return []

    unmatched = [t for t in all_traces if not _is_trace_matched(t)]
    if not unmatched:
        return []

    def _label(t) -> str:
        return (
            getattr(t, "name", None)
            or getattr(t, "filename", None)
            or (getattr(t, "art_id", None) or "<unknown>")
        )

    examples = unmatched[:3]
    bullets = "\n".join(f"- `{_label(t)}`" for t in examples)
    return [
        Recommendation(
            text=(
                "Some traces are not matched to any scenario.\n"
                f"{bullets}"
                + ("\n(…more)" if len(unmatched) > len(examples) else "")
                + "\nRecommendation: add/adjust scenarios so each important trace is explained "
                "(or explicitly link traces to components if you support manual traceability)."
            )
        )
    ]


def _rule_invalid_iml_components(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    dm = new_inst.spec.domain_model

    bad_preds = []
    for p in list(getattr(dm, "state_preds", [])) + list(getattr(dm, "action_preds", [])):
        if getattr(p, "is_iml_valid", None) and str(getattr(p, "is_iml_valid")) == "IMLValidity.INVALID":
            bad_preds.append(p)

    bad_trans = [
        t
        for t in getattr(dm, "transitions", [])
        if getattr(t, "is_iml_valid", None) and str(getattr(t, "is_iml_valid")) == "IMLValidity.INVALID"
    ]

    if not bad_preds and not bad_trans:
        return []

    lines: list[str] = []
    if bad_preds:
        lines.append(
            f"- Invalid predicates: {', '.join(f'`{p.name}`' for p in bad_preds[:6])}{' …' if len(bad_preds) > 6 else ''}"
        )
    if bad_trans:
        lines.append(
            f"- Invalid transitions: {', '.join(f'`{t.name}`' for t in bad_trans[:6])}{' …' if len(bad_trans) > 6 else ''}"
        )

    return [
        Recommendation(
            kind="warning",
            text=(
                "Some model components fail IML/typechecking.\n"
                + "\n".join(lines)
                + "\nRecommendation: fix IML errors first; scenario matching and trace checking won’t be trustworthy otherwise."
            ),
        )
    ]


def _rule_traceability_nudge(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    amap = new_inst.art_map
    if amap is None:
        return []

    n_arts = (
        len(new_inst.art_container.test_traces)
        + len(new_inst.art_container.log_traces)
        + len(new_inst.art_container.doc_ref)
        + len(new_inst.art_container.src_code)
    )
    n_links = sum(len(v) for v in amap.art_to_comp_map.values())

    if n_arts >= 3 and n_links == 0:
        return [
            Recommendation(
                text=(
                    "You have artifacts loaded, but **no traceability links** yet (ArtifactMap is empty).\n"
                    "Recommendation: run trace matching (or add explicit links) so docs/src/traces connect to "
                    "predicates/transitions/scenarios. This makes the synthesized spec explainable."
                )
            )
        ]

    return []


def _rule_state_diff_summary(
    old_inst: StateInstance,
    new_inst: StateInstance,
    failure: ChangeFailure | None,
) -> Iterable[Recommendation]:
    if not _ok(failure):
        return []

    sd: StateDiff | None = new_inst.state_diff
    if sd is None:
        return []

    recs: list[Recommendation] = []

    # 1) If base got better/worse, say that first
    base_out = _outcome(sd.base_status)  # type: ignore[arg-type]
    if base_out in (ComparisonOutcome.IMPROVED, ComparisonOutcome.DECLINED):
        if base_out is ComparisonOutcome.IMPROVED:
            recs.append(
                _rec(
                    f"Domain model base improved ({sd.base_status.before} → {sd.base_status.after}). "
                    "Proceed to checking predicates/transitions and scenario consistency.",
                    kind="info",
                    priority=10,
                )
            )
        else:
            recs.append(
                _rec(
                    f"Domain model base declined ({sd.base_status.before} → {sd.base_status.after}). "
                    "Fix base IML/type issues before trusting any further analysis.",
                    kind="error",
                    priority=0,
                )
            )

    # 2) Predicates
    if _outcome(sd.num_preds_errored) is ComparisonOutcome.IMPROVED:  # type: ignore[arg-type]
        recs.append(
            _rec(
                f"Fewer predicate errors ({sd.num_preds_errored.before} → {sd.num_preds_errored.after}). "
                "Re-run scenario consistency + trace matching to leverage the improved predicate set.",
                kind="info",
                priority=20,
            )
        )
    elif _outcome(sd.num_preds_errored) is ComparisonOutcome.DECLINED:  # type: ignore[arg-type]
        recs.append(
            _rec(
                f"More predicate errors ({sd.num_preds_errored.before} → {sd.num_preds_errored.after}). "
                "Inspect the most recently added/edited predicates and fix their IML first.",
                kind="error",
                priority=5,
            )
        )

    st_err_out = _outcome(sd.num_state_preds_errored)  # type: ignore[arg-type]
    if st_err_out in (ComparisonOutcome.IMPROVED, ComparisonOutcome.DECLINED):
        recs.append(
            _rec(
                f"State-predicate errors changed ({sd.num_state_preds_errored.before} → {sd.num_state_preds_errored.after}). "
                "Focus on `state -> bool` predicates first; they gate most scenario `given` checks.",
                kind="info" if st_err_out is ComparisonOutcome.IMPROVED else "warning",
                priority=23 if st_err_out is ComparisonOutcome.IMPROVED else 9,
            )
        )

    act_err_out = _outcome(sd.num_action_preds_errored)  # type: ignore[arg-type]
    if act_err_out in (ComparisonOutcome.IMPROVED, ComparisonOutcome.DECLINED):
        recs.append(
            _rec(
                f"Action-predicate errors changed ({sd.num_action_preds_errored.before} → {sd.num_action_preds_errored.after}). "
                "Fix `state * action -> bool` predicates; they affect scenario `when` and trace matching.",
                kind="info" if act_err_out is ComparisonOutcome.IMPROVED else "warning",
                priority=24 if act_err_out is ComparisonOutcome.IMPROVED else 10,
            )
        )

    if _outcome(sd.num_preds_matched) is ComparisonOutcome.IMPROVED:  # type: ignore[arg-type]
        recs.append(
            _rec(
                f"More predicates are now matched to artifacts ({sd.num_preds_matched.before} → {sd.num_preds_matched.after}). "
                "Explain these matches in prose and add missing traceability for remaining artifacts.",
                kind="next",
                priority=30,
            )
        )

    # 3) Transitions
    if _outcome(sd.num_trans_valid_logic) is ComparisonOutcome.IMPROVED:  # type: ignore[arg-type]
        recs.append(
            _rec(
                f"More transitions are logic-valid ({sd.num_trans_valid_logic.before} → {sd.num_trans_valid_logic.after}). "
                "Attach these transitions to scenarios’ `then` lists so traces can match end-to-end.",
                kind="next",
                priority=25,
            )
        )
    elif _outcome(sd.num_trans_valid_logic) is ComparisonOutcome.DECLINED:  # type: ignore[arg-type]
        recs.append(
            _rec(
                f"Fewer transitions are logic-valid ({sd.num_trans_valid_logic.before} → {sd.num_trans_valid_logic.after}). "
                "Revert or fix the last transition edit; invalid transitions break scenario execution and matching.",
                kind="error",
                priority=5,
            )
        )

    # 4) Scenarios
    if _outcome(sd.num_sc_missing) is ComparisonOutcome.DECLINED:
        recs.append(
            _rec(
                f"More scenarios are missing components ({sd.num_sc_missing.before} → {sd.num_sc_missing.after}). "
                "Add the referenced predicates/transitions or remove them from those scenarios.",
                kind="error",
                priority=8,
            )
        )
    elif _outcome(sd.num_sc_missing) is ComparisonOutcome.IMPROVED:
        recs.append(
            _rec(
                f"Fewer scenarios are missing components ({sd.num_sc_missing.before} → {sd.num_sc_missing.after}). "
                "Re-check scenario consistency and conflicts now that they are fully specified.",
                kind="next",
                priority=18,
            )
        )

    if _outcome(sd.num_sc_inconsistent) is ComparisonOutcome.DECLINED:
        recs.append(
            _rec(
                f"More scenarios are inconsistent ({sd.num_sc_inconsistent.before} → {sd.num_sc_inconsistent.after}). "
                "Split scenarios or relax contradictory predicate combinations.",
                kind="error",
                priority=12,
            )
        )
    elif _outcome(sd.num_sc_inconsistent) is ComparisonOutcome.IMPROVED:
        recs.append(
            _rec(
                f"Fewer scenarios are inconsistent ({sd.num_sc_inconsistent.before} → {sd.num_sc_inconsistent.after}). "
                "You’re in a good place to match traces and generate human-readable explanations.",
                kind="info",
                priority=22,
            )
        )

    if _outcome(sd.num_sc_conflicted) is ComparisonOutcome.DECLINED:
        recs.append(
            _rec(
                f"More scenario conflicts/overlaps detected ({sd.num_sc_conflicted.before} → {sd.num_sc_conflicted.after}). "
                "Add boundary predicates to make scenarios disjoint, or document intended priority.",
                kind="error",
                priority=15,
            )
        )
    elif _outcome(sd.num_sc_conflicted) is ComparisonOutcome.IMPROVED:
        recs.append(
            _rec(
                f"Fewer scenario conflicts/overlaps ({sd.num_sc_conflicted.before} → {sd.num_sc_conflicted.after}). "
                "Proceed to coverage work—ensure each important trace has at least one matching scenario.",
                kind="next",
                priority=28,
            )
        )

    # 5) Trace coverage
    if _outcome(sd.num_test_traces_logic_good) is ComparisonOutcome.DECLINED:
        recs.append(
            _rec(
                f"More test traces are failing IML validity ({sd.num_test_traces_logic_good.before} → {sd.num_test_traces_logic_good.after}). "
                "Fix the trace encodings (given/when/then) so matching is meaningful.",
                kind="error",
                priority=10,
            )
        )

    if _outcome(sd.num_test_traces_matched) is ComparisonOutcome.IMPROVED:
        recs.append(
            _rec(
                f"More test traces matched scenarios ({sd.num_test_traces_matched.before} → {sd.num_test_traces_matched.after}). "
                "Summarize the matched scenarios as the system’s inferred behavior for those tests.",
                kind="next",
                priority=35,
            )
        )

    if _outcome(sd.num_log_traces_matched) is ComparisonOutcome.IMPROVED:
        recs.append(
            _rec(
                f"More log traces matched scenarios ({sd.num_log_traces_matched.before} → {sd.num_log_traces_matched.after}). "
                "Treat these as in-the-wild confirmation and use them to justify model boundaries.",
                kind="next",
                priority=40,
            )
        )

    if not recs:
        recs.append(
            _rec(
                "No major metric movement detected in the last change. "
                "Add a predicate/transition/scenario (or new traces) to drive measurable progress.",
                kind="info",
                priority=60,
            )
        )

    return recs


#
#   Imandra Inc.
#
#   speclogician/presentation/renderers/state_instance.py
#

from __future__ import annotations

from rich.console import Group, RenderableType
from rich.panel import Panel

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.state_instance import StateInstancePM

from speclogician.presentation.renderers.spec import render_spec
from speclogician.presentation.renderers.state_diff import render_state_diff
from speclogician.presentation.renderers.container import render_artifact_container
from speclogician.presentation.renderers.state_instance_summary import (
    render_state_instance_summary,
)
from speclogician.presentation.renderers.recommendations import render_recommendations

# NEW: trace reachability renderer
from speclogician.presentation.renderers.trace_reachability import render_trace_reachability


def render_state_instance(pm: StateInstancePM, *, ctx: RenderCtx) -> RenderableType:
    summary_panel = render_state_instance_summary(pm.summary, ctx=ctx)

    # Always show recommendations first if present (cheap + high-signal).
    rec_panel: RenderableType | None = None
    if getattr(pm, "recommendations", None) is not None:
        rec_panel = render_recommendations(pm.recommendations, ctx=ctx)

    # Trace reachability (analysis)
    trace_reach_panel: RenderableType | None = None
    if getattr(pm, "trace_reachability", None) is not None:
        trace_reach_panel = render_trace_reachability(pm.trace_reachability, ctx=ctx)

    # Stats-only: recommendations + summary + (optional) trace reachability
    if ctx.show_stats_only:
        blocks: list[RenderableType] = []
        if rec_panel is not None:
            blocks.append(rec_panel)
        blocks.append(summary_panel)
        if trace_reach_panel is not None and (getattr(ctx, "show_analyses", True) or ctx.show_empty_sections):
            blocks.append(
                Panel(
                    trace_reach_panel,
                    title="[bold]Trace reachability[/bold]",
                    border_style="blue",
                    expand=True,
                )
            )
        return Panel(
            Group(*blocks),
            title="[bold]State instance[/bold]",
            border_style="magenta",
            expand=True,
        )

    blocks: list[RenderableType] = []
    if rec_panel is not None:
        blocks.append(rec_panel)

    blocks.append(summary_panel)

    if trace_reach_panel is not None and getattr(ctx, "show_analyses", True):
        blocks.append(
            Panel(
                trace_reach_panel,
                title="[bold]Trace reachability[/bold]",
                border_style="blue",
                expand=True,
            )
        )

    if pm.state_diff is not None and getattr(ctx, "show_diff", True):
        blocks.append(
            Panel(
                render_state_diff(pm.state_diff, ctx=ctx),
                title="[bold]Diff[/bold]",
                border_style="yellow",
                expand=True,
            )
        )

    if pm.spec is not None:
        blocks.append(
            Panel(
                render_spec(pm.spec, ctx=ctx),
                title="[bold]Spec[/bold]",
                border_style="cyan",
                expand=True,
            )
        )

    if pm.artifacts is not None:
        blocks.append(
            Panel(
                render_artifact_container(pm.artifacts, ctx=ctx),
                title="[bold]Artifacts[/bold]",
                border_style="green",
                expand=True,
            )
        )

    return Panel(
        Group(*blocks),
        title="[bold]State instance[/bold]",
        border_style="magenta",
        expand=True,
    )


#
#   Imandra Inc.
#
#   speclogician/presentation/preview/fixtures/state_diff.py
#

from __future__ import annotations

from rich.console import RenderableType

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.state_diff import StateDiffPM, DiffSectionPM, DiffRowPM
from speclogician.state.diff import ComparisonOutcome
from speclogician.presentation.renderers.state_diff import render_state_diff


def make_state_diff_pm(ctx: RenderCtx, variant: str) -> StateDiffPM:
    # keep fixtures deterministic, not dependent on real state
    if variant == "minimal":
        return StateDiffPM(has_changes=False, sections=[])

    if variant == "edge":
        return StateDiffPM(
            has_changes=True,
            sections=[
                DiffSectionPM(
                    title="Domain Model: Base",
                    rows=[
                        DiffRowPM(
                            label="base_status",
                            before="UNKNOWN",
                            after="VALID",
                            outcome=ComparisonOutcome.IMPROVED,
                            message="base_status improved",
                        ),
                        DiffRowPM(
                            label="base_has_action",
                            before=False,
                            after=False,
                            outcome=ComparisonOutcome.NO_CHANGE_BAD,
                            message="still missing action",
                        ),
                        DiffRowPM(
                            label="base_src",
                            before="type state = ...\n" * 3,
                            after="type state = int\ntype action = int\n",
                            outcome=ComparisonOutcome.UNKNOWN,
                            message="multi-line (unknown comparator)",
                        ),
                    ],
                ),
                DiffSectionPM(
                    title="Scenarios",
                    rows=[
                        DiffRowPM(
                            label="num_sc_total",
                            before=3,
                            after=5,
                            outcome=ComparisonOutcome.IMPROVED,
                            message="more scenarios",
                        ),
                        DiffRowPM(
                            label="num_sc_inconsistent",
                            before=0,
                            after=1,
                            outcome=ComparisonOutcome.DECLINED,
                            message="introduced inconsistency",
                        ),
                    ],
                ),
                DiffSectionPM(
                    title="Coverage: Complement",
                    rows=[
                        DiffRowPM(
                            label="scenario_comp_present",
                            before=False,
                            after=True,
                            outcome=ComparisonOutcome.IMPROVED,
                            message="complement now computed",
                        ),
                        DiffRowPM(
                            label="num_comp_regions_total",
                            before=12,
                            after=4,
                            outcome=ComparisonOutcome.IMPROVED,
                            message="fewer uncovered regions",
                        ),
                    ],
                ),
            ],
        )

    return StateDiffPM(
        has_changes=True,
        sections=[
            DiffSectionPM(
                title="Domain Model: Base",
                rows=[
                    DiffRowPM(
                        label="base_status",
                        before="UNKNOWN",
                        after="VALID",
                        outcome=ComparisonOutcome.IMPROVED,
                        message="base status improved",
                    ),
                    DiffRowPM(
                        label="base_has_state",
                        before=False,
                        after=True,
                        outcome=ComparisonOutcome.IMPROVED,
                        message="now has state",
                    ),
                ],
            ),
            DiffSectionPM(
                title="Scenarios",
                rows=[
                    DiffRowPM(
                        label="num_sc_total",
                        before=3,
                        after=5,
                        outcome=ComparisonOutcome.IMPROVED,
                        message="more scenarios",
                    ),
                    DiffRowPM(
                        label="num_sc_inconsistent",
                        before=0,
                        after=1,
                        outcome=ComparisonOutcome.DECLINED,
                        message="inconsistency added",
                    ),
                ],
            ),
            DiffSectionPM(
                title="Coverage: Complement",
                rows=[
                    DiffRowPM(
                        label="scenario_comp_present",
                        before=True,
                        after=True,
                        outcome=ComparisonOutcome.NO_CHANGE_GOOD,
                        message="complement still present",
                    ),
                    DiffRowPM(
                        label="num_comp_regions_total",
                        before=6,
                        after=6,
                        outcome=ComparisonOutcome.NO_CHANGE_BAD,
                        message="coverage not improving yet",
                    ),
                ],
            ),
        ],
    )


def render(pm: StateDiffPM, ctx: RenderCtx) -> RenderableType:
    return render_state_diff(pm, ctx=ctx)



#
#   Imandra Inc.
#
#   speclogician/modeling/trace_reachability_compute.py
#

from __future__ import annotations

from collections import Counter
from math import inf
from typing import Iterable
from dataclasses import dataclass

from speclogician.data.container import ArtifactContainer
from speclogician.data.mapping import ArtifactMap
from speclogician.modeling.reachability import ScenarioReachabilityGraph
from speclogician.modeling.spec import Spec
from speclogician.modeling.trace_reachability import (
    ObservedScenarioHop,
    TraceScenarioAlignment,
    TraceReachabilityAnalysis,
    TraceReachabilitySummary,
)

@dataclass(frozen=True)
class _DPParams:
    reward_edge: float = 2.0
    penalty_no_edge: float = -4.0
    penalty_unmatched_hop: float = -3.0
    penalty_choose_none: float = -0.2


def _dp_choose_step_best(
    *,
    step_matches: list[list[str]],
    reach: ScenarioReachabilityGraph,
    params: _DPParams = _DPParams(),
    allow_none: bool = True,
) -> list[str | None]:
    """
    Choose one label per step to maximize supported hops across the trace.
    Deterministic tie-break: among equal-score solutions, choose the
    lexicographically smallest *entire* sequence.

    We implement tie-break by tracking a per-state "path key" for the best
    solution ending in that state.
    """
    n = len(step_matches)
    if n == 0:
        return []

    # Deterministic state ordering per step
    states: list[list[str | None]] = []
    for cands in step_matches:
        uniq = sorted(set(cands))
        states.append(uniq + ([None] if allow_none else []))

    for i in range(n):
        if not states[i]:
            states[i] = [None]

    def _node_penalty(x: str | None) -> float:
        return params.penalty_choose_none if x is None else 0.0

    def _trans_score(a: str | None, b: str | None) -> float:
        if a is None or b is None:
            return params.penalty_unmatched_hop
        return params.reward_edge if (b in reach.successors(a)) else params.penalty_no_edge

    def _tok(x: str | None) -> tuple[int, str]:
        # None sorts *after* strings
        return (1, "") if x is None else (0, x)

    # dp_score[i][j] = best score up to step i ending in states[i][j]
    # dp_key[i][j]   = lexicographic key for the chosen sequence up to i
    # back[i][j]     = best predecessor index
    dp_score: list[list[float]] = []
    dp_key: list[list[tuple[tuple[int, str], ...]]] = []
    back: list[list[int]] = []

    # init (step 0)
    s0 = []
    k0 = []
    b0 = []
    for x in states[0]:
        s0.append(_node_penalty(x))
        k0.append((_tok(x),))
        b0.append(-1)
    dp_score.append(s0)
    dp_key.append(k0)
    back.append(b0)

    # transitions
    for i in range(1, n):
        cur_scores: list[float] = []
        cur_keys: list[tuple[tuple[int, str], ...]] = []
        cur_back: list[int] = []

        prev_states = states[i - 1]
        cur_states = states[i]

        for j, cur in enumerate(cur_states):
            best_score = -inf
            best_key: tuple[tuple[int, str], ...] | None = None
            best_k = 0

            for k, prev in enumerate(prev_states):
                score = dp_score[i - 1][k] + _trans_score(prev, cur) + _node_penalty(cur)
                key = dp_key[i - 1][k] + (_tok(cur),)

                if score > best_score:
                    best_score = score
                    best_key = key
                    best_k = k
                elif score == best_score:
                    assert best_key is not None
                    # global lexicographic tie-break on the full sequence key
                    if key < best_key:
                        best_key = key
                        best_k = k

            cur_scores.append(best_score)
            cur_keys.append(best_key if best_key is not None else ())
            cur_back.append(best_k)

        dp_score.append(cur_scores)
        dp_key.append(cur_keys)
        back.append(cur_back)

    # pick best final state by (score, key)
    last_scores = dp_score[-1]
    last_keys = dp_key[-1]

    best_j = 0
    for j in range(1, len(states[-1])):
        if last_scores[j] > last_scores[best_j]:
            best_j = j
        elif last_scores[j] == last_scores[best_j]:
            if last_keys[j] < last_keys[best_j]:
                best_j = j

    # backtrack
    out: list[str | None] = [None] * n
    j = best_j
    for i in range(n - 1, -1, -1):
        out[i] = states[i][j]
        j = back[i][j]
        if j < 0:
            break

    return out

def compute_trace_reachability_analysis(
    *,
    spec: Spec,
    art_cont: ArtifactContainer,
    art_map: ArtifactMap,
    reach: ScenarioReachabilityGraph,
    include_traces: Iterable[str] | None = None,
    cap_step_matches: int | None = None,
    include_shortest_path: bool = True,
) -> TraceReachabilityAnalysis:
    """
    Compute TraceReachabilityAnalysis based on:
      - reachability graph between scenarios (reach)
      - step-level mapping in art_map using trace.step_ref(i)
 synchronously, purely in-memory.

    Assumptions:
      - Scenario IDs used in the graph are scenario.comp_id.
      - Step-level mapping uses trace.step_ref(i) -> [scenario.comp_id, ...]
    """

    scenario_ids = [sc.comp_id for sc in spec.scenarios]
    scenario_set = set(scenario_ids)

    # Choose traces
    traces = art_cont.all_traces()
    if include_traces is not None:
        want = set(include_traces)
        traces = [t for t in traces if t.art_id in want]

    per_trace: dict[str, TraceScenarioAlignment] = {}

    # Summary accumulators
    total_steps = 0
    total_hops = 0
    supported_hops = 0
    unsupported_hops = 0

    scenarios_hit: set[str] = set()
    unsupported_edge_counts: Counter[tuple[str, str]] = Counter()
    frontier_end_counts: Counter[str] = Counter()

    for tr in traces:
        if tr.art_id is None:
            continue

        n = len(tr.steps)
        total_steps += n
        total_hops += max(0, n - 1)

        step_matches: list[list[str]] = []
        step_best: list[str | None] = []

        # ---- step matches ----
        step_matches: list[list[str]] = []

        for i in range(n):
            ref = tr.step_ref(i)
            cands = list(art_map.get_comp_for_artifact(ref))
            cands = [c for c in cands if c in scenario_set]   # filter stale
            if cap_step_matches is not None:
                cands = cands[:cap_step_matches]
            step_matches.append(cands)
            for c in cands:
                scenarios_hit.add(c)

        # NEW: choose globally-optimal step_best
        step_best = _dp_choose_step_best(step_matches=step_matches, reach=reach)

        # ---- hops ----
        observed_hops: list[ObservedScenarioHop] = []
        supported_local = 0
        unsupported_local = 0

        for i in range(n - 1):
            src = step_best[i]
            dst = step_best[i + 1]

            hop = ObservedScenarioHop(
                trace_id=tr.art_id,
                step_from=i,
                step_to=i + 1,
                src=src or "",
                dst=dst or "",
                supported_by_graph=False,
                shortest_path=None,
                reason=None,
            )

            if src is None or dst is None:
                hop.supported_by_graph = False
                hop.reason = "unmatched_step"
                unsupported_local += 1
            else:
                if dst in reach.successors(src):
                    hop.supported_by_graph = True
                    supported_local += 1
                else:
                    hop.supported_by_graph = False
                    hop.reason = "no_graph_edge"
                    unsupported_local += 1
                    unsupported_edge_counts[(src, dst)] += 1

                    if include_shortest_path:
                        path = reach.shortest_path(src, dst)
                        hop.shortest_path = path
                        if path is None:
                            hop.reason = "no_path"

            observed_hops.append(hop)

        # “frontier ending” = last matched label (useful for UI triage)
        for j in range(n - 1, -1, -1):
            if step_best[j] is not None:
                frontier_end_counts[step_best[j]] += 1
                break

        supported_hops += supported_local
        unsupported_hops += unsupported_local

        per_trace[tr.art_id] = TraceScenarioAlignment(
            trace_id=tr.art_id,
            step_matches=step_matches,
            step_best=step_best,
            observed_hops=observed_hops,
            num_steps=n,
            num_matched_steps=sum(1 for x in step_best if x is not None),
            num_supported_hops=supported_local,
            num_unsupported_hops=unsupported_local,
        )

    scenarios_unhit = sorted(list(scenario_set - scenarios_hit))
    scenarios_hit_list = sorted(list(scenarios_hit))

    top_unsupported_edges = [
        (src, dst, cnt) for (src, dst), cnt in unsupported_edge_counts.most_common(25)
    ]
    top_frontier_trace_endings = [
        (sc, cnt) for sc, cnt in frontier_end_counts.most_common(25)
    ]

    summary = TraceReachabilitySummary(
        total_traces=len(per_trace),
        total_steps=total_steps,
        total_hops=total_hops,
        supported_hops=supported_hops,
        unsupported_hops=unsupported_hops,
        scenarios_hit=scenarios_hit_list,
        scenarios_unhit=scenarios_unhit,
        top_unsupported_edges=top_unsupported_edges,
        top_frontier_trace_endings=top_frontier_trace_endings,
    )

    return TraceReachabilityAnalysis(
        per_trace=per_trace,
        summary=summary,
        # fingerprints: fill if you have stable hashing utilities
        spec_fp=None,
        artifacts_fp=None,
        mapping_fp=None,
    )


#
#   Imandra Inc.
#
#   speclogician/presentation/models/state_instance_summary.py
#

from __future__ import annotations

from pydantic import BaseModel
from typing import Optional


class StateInstanceSummaryPM(BaseModel):
    # -----------------
    # metadata
    # -----------------
    created_at: Optional[str] = None

    # -----------------
    # changes
    # -----------------
    num_changes: int = 0

    # -----------------
    # Domain model (base)
    # -----------------
    base_status: Optional[str] = None
    base_has_state: Optional[bool] = None
    base_has_action: Optional[bool] = None

    # -----------------
    # Domain model (state predicates)
    # -----------------
    num_state_preds_total: Optional[int] = None
    num_state_preds_matched: Optional[int] = None
    num_state_preds_valid_logic: Optional[int] = None
    num_state_preds_errored: Optional[int] = None

    # -----------------
    # Domain model (action predicates)
    # -----------------
    num_action_preds_total: Optional[int] = None
    num_action_preds_matched: Optional[int] = None
    num_action_preds_valid_logic: Optional[int] = None
    num_action_preds_errored: Optional[int] = None

    # -----------------
    # Domain model (predicate totals)
    # -----------------
    num_preds_total: Optional[int] = None
    num_preds_matched: Optional[int] = None
    num_preds_valid_logic: Optional[int] = None
    num_preds_errored: Optional[int] = None

    # -----------------
    # Domain model (transitions)
    # -----------------
    num_trans_total: Optional[int] = None
    num_trans_matched: Optional[int] = None
    num_trans_valid_logic: Optional[int] = None
    num_trans_errored: Optional[int] = None

    # -----------------
    # Scenarios
    # -----------------
    num_sc_total: Optional[int] = None
    num_sc_missing: Optional[int] = None
    num_sc_matched: Optional[int] = None
    num_sc_inconsistent: Optional[int] = None

    # -----------------
    # Conflicts (spec-level)
    # -----------------
    num_sc_conflicted: Optional[int] = None
    num_sc_overlap: Optional[int] = None
    num_sc_consumed: Optional[int] = None

    # -----------------
    # Test traces
    # -----------------
    num_test_traces_total: Optional[int] = None
    num_test_traces_matched: Optional[int] = None
    num_test_traces_logic_good: Optional[int] = None

    # -----------------
    # Log traces
    # -----------------
    num_log_traces_total: Optional[int] = None
    num_log_traces_matched: Optional[int] = None
    num_log_traces_logic_good: Optional[int] = None

    # -----------------
    # Source code artifacts
    # -----------------
    num_src_code_arts_total: Optional[int] = None
    num_src_code_arts_matched: Optional[int] = None

    # -----------------
    # Doc artifacts
    # -----------------
    num_doc_arts_total: Optional[int] = None
    num_doc_arts_matched: Optional[int] = None


#
#   Imandra Inc.
#
#   speclogician/logic/lib/complement.py
#

from __future__ import annotations

import asyncio
from typing import Optional

from speclogician.modeling.domain import BaseStatus
from speclogician.modeling.complement import ScenarioComplement
from speclogician.modeling.spec import Spec

from ..api.decomp import decomp_complement_async


def _is_iml_invalid(x: object) -> bool:
    """
    Best-effort check for IML invalidity without hard-coding exact enum imports.
    Works with:
      - IMLValidity.INVALID enum
      - string/repr containing "IMLValidity.INVALID"
      - literal "invalid"
    """
    v = getattr(x, "is_iml_valid", None)
    if v is None:
        return False

    # enum-style
    try:
        if getattr(v, "name", None) == "INVALID":
            return True
    except Exception:
        pass

    # string-style
    s = str(v)
    return (s == "invalid") or ("IMLValidity.INVALID" in s)


def _spec_is_decomposable_for_complement(spec: Spec) -> tuple[bool, str]:
    dm = spec.domain_model

    # 1) base must be good
    if getattr(dm, "base_status", None) != BaseStatus.VALID:
        return False, f"base_status={getattr(dm, 'base_status', None)}"

    # 2) must have scenarios
    scs = list(getattr(spec, "scenarios", []) or [])
    if len(scs) == 0:
        return False, "no_scenarios"

    # 3) scenarios should be internally consistent (avoid wasting time)
    if any(getattr(sc, "all_preds_consistent", True) is False for sc in scs):
        return False, "has_inconsistent_scenarios"

    # 4) avoid complement if any predicate/transition is IML-invalid
    bad: list[tuple[str, str]] = []

    for p in list(getattr(dm, "state_preds", [])) + list(getattr(dm, "action_preds", [])):
        if _is_iml_invalid(p):
            bad.append(("pred", getattr(p, "name", "<unknown>")))

    for t in list(getattr(dm, "transitions", [])):
        if _is_iml_invalid(t):
            bad.append(("trans", getattr(t, "name", "<unknown>")))

    if bad:
        return False, f"iml_invalid={bad[:3]}{'...' if len(bad) > 3 else ''}"

    return True, "ok"


def maybe_update_spec_complement(spec: Spec) -> Optional[ScenarioComplement]:
    """
    Policy-level helper (SYNC):

      - decides whether it is *appropriate* to compute complement
      - if not appropriate, clears stale complement
      - if appropriate, calls ImandraX via API-layer decomp and stores result

    NOTE:
      - This is intentionally sync to fit the state pipeline.
      - It runs the async decomp using asyncio.run().
      - If called inside an existing event loop, it will gracefully skip and
        leave complement cleared (to avoid stale info).
    """
    ok, _why = _spec_is_decomposable_for_complement(spec)
    if not ok:
        # Prevent stale coverage info from hanging around.
        spec.scenario_comp = None
        return None

    try:
        comp = asyncio.run(decomp_complement_async(spec))
    except RuntimeError:
        # "asyncio.run() cannot be called from a running event loop"
        # This should be rare in CLI, but can happen in notebooks/tests.
        # Prefer "clear rather than stale".
        spec.scenario_comp = None
        return None

    spec.scenario_comp = comp
    return comp




#
#   Imandra Inc.
#
#   speclogician/commands/runner.py
#

from __future__ import annotations

import json
import typer
from enum import Enum
from datetime import datetime
from typing import NoReturn, Any
from rich.console import Console

from speclogician.state.change import StateChange
from speclogician.state.state import State
from speclogician.state.change_result import (
    ProcessChangeResult,
    ProcessChangeError,
    ProcessChangeSuccess,
    ProcessBatchResult,
    ProcessBatchError,
    ProcessBatchSuccess,
)

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.builders.state_instance import build_state_instance_pm
from speclogician.presentation.builders.state_diff import build_state_diff_pm

from speclogician.presentation.renderers.state_instance import render_state_instance
from speclogician.presentation.renderers.state_diff import render_state_diff  # assuming you have it

console = Console()


def _dump_json(obj: Any) -> str:
    def _default(o: Any):
        if hasattr(o, "model_dump"):
            return o.model_dump(mode="json")
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, datetime):
            return o.isoformat()
        raise TypeError(f"Object of type {type(o).__name__} is not JSON serializable")

    return json.dumps(
        obj,
        default=_default,
        indent=2,
        sort_keys=True,
    )

def run_change_and_render(
    st: State,
    change: StateChange,
    *,
    json_only: bool,
) -> NoReturn:
    """
    Unified runner for CLI commands.

    - json_only=True: emits JSON PM payload to stdout (stable, parseable).
    - json_only=False: pretty-prints PMs via Rich renderers.
    """

    # Always take the object path so we can drive presentation from PMs.
    result_obj: ProcessChangeResult = st.process_change(change, json_only=json_only)
    if result_obj["ok"]:
        st.save()

    # -------------------------
    # Error branch
    # -------------------------
    if not result_obj["ok"]:
        err_res: ProcessChangeError = result_obj

        # Build ctx from the old state if possible
        old_state = err_res["old_state"]
        ctx = RenderCtx(
            art_cont=old_state.art_container,
            art_map=old_state.art_map,
        )

        old_pm = build_state_instance_pm(old_state, ctx=ctx)

        if json_only:
            payload = {
                "ok": False,
                "stage": err_res["stage"],
                "change": err_res["change"].model_dump(mode="json"),
                "error": {
                    "type": type(err_res["error"]).__name__,
                    "message": str(err_res["error"]),
                },
                "old_state": old_pm.model_dump(mode="json"),
            }
            typer.echo(_dump_json(payload))
            raise typer.Exit(code=1)

        # human output
        console.print(f"[bold red]Error[/bold red] stage={err_res['stage']}")
        console.print(f"[red]{type(err_res['error']).__name__}: {err_res['error']}[/red]\n")
        console.print("[bold]Old state[/bold]")
        console.print(render_state_instance(old_pm, ctx=ctx))
        raise typer.Exit(code=1)

    # -------------------------
    # Success branch
    # -------------------------
    ok_res: ProcessChangeSuccess = result_obj

    old_state = ok_res["old_state"]
    new_state = ok_res["new_state"]
    sd = ok_res["state_diff"]

    # Prefer new-state context for presentation (has freshest art_map/art_cont)
    ctx = RenderCtx(
        art_cont=new_state.art_container,
        art_map=new_state.art_map,
    )

    old_pm = build_state_instance_pm(old_state, ctx=ctx)
    new_pm = build_state_instance_pm(new_state, ctx=ctx)
    diff_pm = build_state_diff_pm(sd, ctx=ctx)

    if json_only:
        payload = {
            "ok": True,
            "stage": "ok",
            "change": ok_res["change"].model_dump(mode="json"),
            "old_state": old_pm.model_dump(mode="json"),
            "new_state": new_pm.model_dump(mode="json"),
            "state_diff": diff_pm.model_dump(mode="json"),
            "new_num_instances": ok_res["new_num_instances"],
        }
        typer.echo(_dump_json(payload))
        raise typer.Exit(code=0)

    # human output
    console.print("[bold]Old state[/bold]")
    console.print(render_state_instance(old_pm, ctx=ctx))
    console.print()

    console.print("[bold]New state[/bold]")
    console.print(render_state_instance(new_pm, ctx=ctx))
    console.print()

    console.print("[bold]Diff[/bold]")
    console.print(render_state_diff(diff_pm, ctx=ctx))
    console.print()

    raise typer.Exit(code=0)


def run_batch_and_render(
    st: State,
    changes: list[StateChange],
    *,
    json_only: bool,
) -> NoReturn:
    """
    Unified runner for batch CLI commands.
    """

    result_obj: ProcessBatchResult = st.process_batch(changes, json_only=json_only)
    if result_obj["ok"]:
        st.save()

    # -------------------------
    # Error branch
    # -------------------------
    if not result_obj["ok"]:
        err_res: ProcessBatchError = result_obj

        old_state = err_res["old_state"]
        rctx = RenderCtx(
            art_cont=old_state.art_container,
            art_map=old_state.art_map,
        )

        old_pm = build_state_instance_pm(old_state, ctx=rctx)

        if json_only:
            payload = {
                "ok": False,
                "stage": err_res["stage"],
                "change_index": err_res["change_index"],
                "changes": [c.model_dump(mode="json") for c in err_res["changes"]],
                "error": {
                    "type": type(err_res["error"]).__name__,
                    "message": str(err_res["error"]),
                },
                "old_state": old_pm.model_dump(mode="json"),
            }
            typer.echo(_dump_json(payload))
            raise typer.Exit(code=1)

        # human output
        idx = err_res["change_index"]
        idx_str = f" (change #{idx})" if idx >= 0 else ""
        console.print(f"[bold red]Batch Error[/bold red] stage={err_res['stage']}{idx_str}")
        console.print(f"[red]{type(err_res['error']).__name__}: {err_res['error']}[/red]\n")
        console.print("[bold]Old state[/bold]")
        console.print(render_state_instance(old_pm, ctx=rctx))
        raise typer.Exit(code=1)

    # -------------------------
    # Success branch
    # -------------------------
    ok_res: ProcessBatchSuccess = result_obj

    old_state = ok_res["old_state"]
    new_state = ok_res["new_state"]
    sd = ok_res["state_diff"]

    rctx = RenderCtx(
        art_cont=new_state.art_container,
        art_map=new_state.art_map,
    )

    old_pm = build_state_instance_pm(old_state, ctx=rctx)
    new_pm = build_state_instance_pm(new_state, ctx=rctx)
    diff_pm = build_state_diff_pm(sd, ctx=rctx)

    if json_only:
        payload = {
            "ok": True,
            "stage": "ok",
            "changes": [c.model_dump(mode="json") for c in ok_res["changes"]],
            "old_state": old_pm.model_dump(mode="json"),
            "new_state": new_pm.model_dump(mode="json"),
            "state_diff": diff_pm.model_dump(mode="json"),
            "new_num_instances": ok_res["new_num_instances"],
        }
        typer.echo(_dump_json(payload))
        raise typer.Exit(code=0)

    # human output
    console.print(f"[bold]Batch applied ({len(ok_res['changes'])} changes)[/bold]\n")

    console.print("[bold]Old state[/bold]")
    console.print(render_state_instance(old_pm, ctx=rctx))
    console.print()

    console.print("[bold]New state[/bold]")
    console.print(render_state_instance(new_pm, ctx=rctx))
    console.print()

    console.print("[bold]Diff[/bold]")
    console.print(render_state_diff(diff_pm, ctx=rctx))
    console.print()

    raise typer.Exit(code=0)


#
#   Imandra Inc.
#
#   speclogician/presentation/models/scenario.py
#

from __future__ import annotations

from typing import Literal
from pydantic import BaseModel, Field


class ScenarioStatusPM(BaseModel):
    kind: Literal["valid", "missing", "inconsistent"]
    missing_preds: list[str] = Field(default_factory=list)
    missing_trans: list[str] = Field(default_factory=list)


class ScenarioPM(BaseModel):
    comp_id: str
    name: str

    status: ScenarioStatusPM

    given: list[str] = Field(default_factory=list)
    when: list[str] = Field(default_factory=list)
    then: list[str] = Field(default_factory=list)

    given_preds_consistent: bool = True
    when_preds_consistent: bool = True
    all_preds_consistent: bool = True


#
#   Imandra Inc.
#
#   speclogician/commands/reach.py
#

from __future__ import annotations
from typing import Any

import json
import time
import typer

from speclogician.utils import console
from speclogician.logic.lib.reachability import ScenarioReachabilityGraph

from .utils import get_state

app = typer.Typer(help="Scenario reachability analysis (ImandraX-backed compute + graph queries).")

def _emit_json(payload: dict[str, Any]) -> None:
    typer.echo(json.dumps(payload, indent=2, sort_keys=True, default=str))

def check_graph(g:ScenarioReachabilityGraph|None, json_only:bool) -> ScenarioReachabilityGraph:
    if g is None:
        if json_only:
            _emit_json({"ok": False, "notes": "Scenario reachability not computed"})
        else:
            console.print("[dim]No reachability graph computed yet. Run: speclogician reach compute[/dim]")
        raise typer.Exit(0)
    return g

@app.command("summary")
def summary_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json")
) -> None:

    state = get_state(ctx)
    spec = state.curr_state().spec
    g = spec.scenario_reachability

    g = check_graph(g, json_only)

    if json_only:
        _emit_json(
            {
                "ok": True,
                "nodes": len(g.nodes),
                "edges": len(g.edges),
                "initials": g.initials,
                "reachable_from_initials": g.reachable_from_initials,
                "unreachable": g.unreachable,
            }
        )
        raise typer.Exit(0)

    console.print(f"[bold]Reachability graph[/bold] nodes={len(g.nodes)} edges={len(g.edges)}")
    if g.initials:
        console.print(f"[dim]initials[/dim] {', '.join(g.initials)}")
        console.print(f"[dim]unreachable[/dim] {len(g.unreachable)}")


@app.command("unreachable")
def unreachable_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json")
) -> None:
    
    state = get_state(ctx)

    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    if json_only:
        _emit_json({"unreachable": g.unreachable})
    else:
        for n in g.unreachable:
            console.print(n)

@app.command("succ")
def succ_cmd(
    ctx: typer.Context,
    name: str,
    json_only: bool = typer.Option(False, "--json")
) -> None:

    state = get_state(ctx)
    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    for n in g.successors(name):
        console.print(n)

@app.command("pred")
def pred_cmd(
    ctx: typer.Context,
    name: str,
    json_only: bool = typer.Option(False, "--json")
) -> None:
    state = get_state(ctx)

    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    for n in g.predecessors(name):
        console.print(n)


@app.command("path")
def path_cmd(
    ctx: typer.Context,
    frm: str,
    to: str, 
    json_only: bool = typer.Option(False, "--json")
) -> None:
    
    state = get_state(ctx)
    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    path = g.shortest_path(frm, to)  # helper on graph
    if json_only:
        _emit_json({"path": path})
    else:
        console.print(" -> ".join(path) if path else "[dim]no path[/dim]")


@app.command("explain-edge")
def explain_edge_cmd(
    ctx: typer.Context,
    src: str,
    dst: str,
    no_witness: bool = typer.Option(False, "--no-witness"),
    json_only: bool = typer.Option(False, "--json"),
) -> None:

    state = get_state(ctx)
    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    ex = g.explain_edge(src, dst, include_witness_src=not no_witness)  # helper on graph

    if json_only:
        _emit_json(ex)
        raise typer.Exit(0)

    if not ex.get("reachable"):
        console.print(f"[red]not reachable[/red] {src} -> {dst}")
        notes = ex.get("notes")
        if notes:
            console.print(f"[dim]{notes}[/dim]")
        raise typer.Exit(1)

    console.print(f"[green]reachable[/green] {src} -> {dst}")
    if ex.get("witness_fp"):
        console.print(f"[dim]witness_fp[/dim] {ex['witness_fp']}")
    if ex.get("witness_src"):
        console.print(ex["witness_src"])


@app.command("frontier")
def frontier_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json")
) -> None:
    
    state = get_state(ctx)
    g = state.curr_state().spec.scenario_reachability

    g = check_graph(g, json_only)

    gaps = g.frontier_gaps()

    if json_only:
        _emit_json({"frontier_gaps": [{"src": s, "missing_reason": r} for (s, r) in gaps]})
        raise typer.Exit(0)

    for s, r in gaps:
        console.print(f"{s}  [dim]{r}[/dim]")


#
#   Imandra Inc.
#
#   speclogician/tui/stats.py
#

from typing import Any
from textual.screen import Screen
from textual.app import ComposeResult
from textual.widgets import Header, Footer
from textual.containers import Container
from textual_plotext import PlotextPlot

from ..state.state_stats import StatsCalculator


class StatsScreen(Screen[Any]):
    CSS = """
    #body { height: 1fr; }

    /* 2x2 grid */
    #grid {
        height: 1fr;
        layout: grid;
        grid-size: 2 2;
        grid-gutter: 1 1;
    }

    PlotextPlot {
        height: 1fr;
        width: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield Header()
        with Container(id="body"):
            with Container(id="grid"):
                self.plot_model = PlotextPlot()
                self.plot_scenarios = PlotextPlot()
                self.plot_artifacts = PlotextPlot()
                self.plot_ratios = PlotextPlot()

                yield self.plot_model
                yield self.plot_scenarios
                yield self.plot_artifacts
                yield self.plot_ratios
        yield Footer()

    def on_mount(self) -> None:
        # If you want live updating while running:
        # self.set_interval(0.5, self.refresh_plot)
        self.refresh_plot()

    @staticmethod
    def _safe_ratio_series(num: list[int], den: list[int]) -> list[float]:
        out: list[float] = []
        for n, d in zip(num, den):
            out.append(n / (d if d else 1))
        return out

    def refresh_plot(self) -> None:

        v = StatsCalculator.stats(self.app.state.instances)

        # oldest -> newest already
        xs = list(range(len(v["created_at"])))

        # -------------------------
        # 1) Model growth & usage
        # -------------------------
        plt = self.plot_model.plt
        plt.clear_data()
        plt.title("Model growth & usage")
        plt.xlabel("Instance (oldest → newest)")
        plt.ylabel("count")
        plt.plot(xs, v["num_preds_total"])
        plt.plot(xs, v["num_preds_matched"])
        plt.plot(xs, v["num_trans_total"])
        plt.scatter(xs, v["num_preds_matched"])
        self.plot_model.refresh()

        # -------------------------
        # 2) Scenario health
        # -------------------------
        plt = self.plot_scenarios.plt
        plt.clear_data()
        plt.title("Scenario health")
        plt.xlabel("Instance (oldest → newest)")
        plt.ylabel("count")
        plt.plot(xs, v["num_sc_total"])
        plt.plot(xs, v["num_sc_matched"])
        plt.plot(xs, v["num_sc_missing"])
        plt.plot(xs, v["num_sc_inconsistent"])
        plt.plot(xs, v["num_sc_conflicted"])
        self.plot_scenarios.refresh()

        # -------------------------
        # 3) Artifact pipeline quality
        # -------------------------
        plt = self.plot_artifacts.plt
        plt.clear_data()
        plt.title("Artifacts: total → logic-good → matched")
        plt.xlabel("Instance (oldest → newest)")
        plt.ylabel("count")
        plt.plot(xs, v["num_traces_total"])
        plt.plot(xs, v["num_traces_logic_good"])
        plt.plot(xs, v["num_traces_matched"])
        plt.scatter(xs, v["num_traces_matched"])
        self.plot_artifacts.refresh()

        # -------------------------
        # 4) Coverage ratios
        # -------------------------
        plt = self.plot_ratios.plt
        plt.clear_data()
        plt.title("Coverage ratios")
        plt.xlabel("Instance (oldest → newest)")
        plt.ylabel("ratio (0..1)")
        preds_ratio = self._safe_ratio_series(v["num_preds_matched"], v["num_preds_total"])
        traces_ratio = self._safe_ratio_series(v["num_traces_matched"], v["num_traces_total"])
        plt.plot(xs, preds_ratio)
        plt.plot(xs, traces_ratio)
        plt.scatter(xs, preds_ratio)
        self.plot_ratios.refresh()


#
#   Imandra Inc.
#
#   speclogician/presentation/models/scenario_complement.py
#

from __future__ import annotations

from typing import Literal
from pydantic import BaseModel


class ScenarioComplementRegionPM(BaseModel):
    """
    Lightweight representation of a single region.

    We store:
      - fp: stable-ish short fingerprint (from modeling/complement.py)
      - invariant/model/model_eval/constraints: optional strings for display
    """
    fp: str

    num_constraints: int = 0
    invariant: str | None = None
    model: str | None = None
    model_eval: str | None = None

    # optional: include a few constraints for preview
    constraints_preview: list[str] = []


class ScenarioComplementPM(BaseModel):
    kind: Literal["scenario_complement"] = "scenario_complement"

    count_regions: int = 0

    # a small sample for UI
    regions_preview: list[ScenarioComplementRegionPM] = []

    regions_all: list[ScenarioComplementRegionPM] | None = None



#
#   Imandra Inc.
#
#   speclogician/presentation/models/state_instance.py
#

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel

from speclogician.presentation.models.state_instance_summary import StateInstanceSummaryPM
from speclogician.presentation.models.spec import SpecPM
from speclogician.presentation.models.container import ArtifactContainerPM
from speclogician.presentation.models.state_diff import StateDiffPM
from speclogician.presentation.models.recommendations import RecommendationsPM

# NEW: trace reachability presentation model
from speclogician.presentation.models.trace_reachability import TraceReachabilityPM


class StateInstancePM(BaseModel):
    kind: Literal["state_instance"] = "state_instance"

    summary: StateInstanceSummaryPM

    # Optional heavy sections
    spec: SpecPM | None = None
    artifacts: ArtifactContainerPM | None = None
    state_diff: StateDiffPM | None = None

    # Analyses
    trace_reachability: TraceReachabilityPM | None = None

    # Recommendations (computed after change/diff)
    recommendations: RecommendationsPM


#
#   Imandra Inc.
#
#   speclogician/logic/main.py
#

from __future__ import annotations

from ..modeling.spec import Spec
from ..data.container import ArtifactContainer
from ..data.mapping import ArtifactMap

from ..state.change import (
    StateChange,
    DomainModelBaseEdit,
    PredicateAdd,
    PredicateEdit,
    PredicateRemove,
    TransitionAdd,
    TransitionEdit,
    TransitionRemove,
    ScenarioAdd,
    ScenarioEdit,
    ScenarioRemove,
    AddTestTrace,
    EditTestTrace,
    AddLogTrace,
    EditLogTrace
)

from .lib.domain_model import check_domain_model
from .lib.predicates import check_predicates
from .lib.transitions import check_transitions
from .lib.scenarios import check_scenarios
from .lib.traces import check_traces
from .lib.complement import maybe_update_spec_complement
from .lib import update_reachability_graph

from ..utils import console


def _hdr(title: str, subtitle: str | None = None, json_only: bool = False) -> None:
    if json_only:
        return
    from rich.panel import Panel
    from rich.text import Text

    console.print(
        Panel(
            Text(subtitle or ""),
            title=f"[bold]Logician[/bold] · {title}",
            border_style="cyan",
        )
    )


def _note(msg: str, json_only: bool = False) -> None:
    if not json_only:
        console.print(msg)

def analyze_change(
    spec: Spec,
    art_cont: ArtifactContainer,
    art_map: ArtifactMap,
    change: StateChange,
    json_only: bool = False,
) -> None:
    """
    Top-level entry for logical analysis post-change event.

    Only performs logic checks when relevant.
    All other change types are treated as no-ops here.
    """

    # ----------------------------
    # Domain base model changed
    # ----------------------------
    if isinstance(change, DomainModelBaseEdit):
        _hdr("DomainModelBaseEdit", "Rechecking domain model and all dependent logic", json_only=json_only)
        check_domain_model(spec, json_only=json_only)
        check_predicates(spec, json_only=json_only)
        check_transitions(spec, json_only=json_only)
        check_scenarios(spec, json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        _note("[green]✓[/green] Domain analysis complete", json_only=json_only)
        return

    # ----------------------------
    # Predicates changed
    # ----------------------------
    if isinstance(change, (PredicateAdd, PredicateEdit)):
        _hdr("Predicate change", f"Predicate: [bold]{change.pred_name}[/bold]", json_only=json_only)
        check_predicates(spec, names=[change.pred_name], json_only=json_only)
        check_scenarios(spec, predicates=[change.pred_name], json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)


        _note("[green]✓[/green] Predicate analysis complete", json_only=json_only)
        return

    if isinstance(change, PredicateRemove):
        _hdr("Predicate removed", f"[bold]{change.pred_name}[/bold]", json_only=json_only)
        check_scenarios(spec, predicates=[change.pred_name], json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        _note("[green]✓[/green] Scenario refresh complete", json_only=json_only)
        return

    # ----------------------------
    # Transitions changed
    # ----------------------------
    if isinstance(change, (TransitionAdd, TransitionEdit)):
        _hdr("Transition change", f"Transition: [bold]{change.trans_name}[/bold]", json_only=json_only)
        check_transitions(spec, names=[change.trans_name], json_only=json_only)
        check_scenarios(spec, transitions=[change.trans_name], json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        _note("[green]✓[/green] Transition analysis complete", json_only=json_only)
        return

    if isinstance(change, TransitionRemove):
        _hdr("Transition removed", f"[bold]{change.trans_name}[/bold]", json_only=json_only)
        check_scenarios(spec, transitions=[change.trans_name], json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        _note("[green]✓[/green] Scenario refresh complete", json_only=json_only)
        return

    # ----------------------------
    # Scenarios changed
    # ----------------------------
    if isinstance(change, (ScenarioAdd, ScenarioEdit)):
        _hdr("Scenario change", f"Scenario: [bold]{change.scenario_name}[/bold]", json_only=json_only)
        check_scenarios(spec, names=[change.scenario_name], json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)


        _note("[green]✓[/green] Scenario analysis complete", json_only=json_only)
        return

    if isinstance(change, ScenarioRemove):
        _hdr("Scenario removed", "Refreshing scenarios, predicates, and traces", json_only=json_only)
        check_scenarios(spec, json_only=json_only)
        check_predicates(spec, json_only=json_only)
        check_traces(spec=spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)

        try:
            maybe_update_spec_complement(spec)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

        try:
            update_reachability_graph(spec, json_only=json_only)
        except Exception as e:
            _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)


        _note("[green]✓[/green] Full refresh complete", json_only=json_only)
        return

    # ----------------------------
    # Trace artifacts changed
    # ----------------------------
    if isinstance(change, AddTestTrace):
        if not art_cont.test_traces:
            return

        tt = art_cont.test_traces[-1]
        # best-effort sanity check; no-op if mismatch
        if change.art_id is not None and tt.art_id != change.art_id:
            pass

        assert tt.art_id is not None

        check_traces(
            spec=spec,
            art_cont=art_cont,
            art_map=art_map,
            traces=[tt.art_id],
            json_only=json_only,
        )

        _note("[green]✓[/green] Trace analysis complete", json_only=json_only)
        return

    if isinstance(change, AddLogTrace):
        if not art_cont.log_traces:
            return

        lt = art_cont.log_traces[-1]
        if change.art_id is not None and lt.art_id != change.art_id:
            pass

        assert lt.art_id is not None

        check_traces(
            spec=spec,
            art_cont=art_cont,
            art_map=art_map,
            traces=[lt.art_id],
            json_only=json_only,
        )

        _note("[green]✓[/green] Trace analysis complete", json_only=json_only)
        return

    if isinstance(change, (EditTestTrace, EditLogTrace)):
        check_traces(
            spec=spec,
            art_cont=art_cont,
            art_map=art_map,
            traces=[change.art_id],
            json_only=json_only,
        )

        _note("[green]✓[/green] Trace analysis complete", json_only=json_only)
        return

    # ----------------------------
    # Explicit no-op branch
    # ----------------------------
    # All remaining change types (data artifacts, mapping edits, etc.)
    # are intentionally ignored here and handled elsewhere.
    return


def analyze_batch(
    spec: Spec,
    art_cont: ArtifactContainer,
    art_map: ArtifactMap,
    changes: list[StateChange],
    json_only: bool = False,
) -> None:
    """
    Batch-aware analysis: compute the union scope of all changes,
    then run each check function once with the merged scope.
    """

    # If any change is a DomainModelBaseEdit, do a full recheck
    if any(isinstance(c, DomainModelBaseEdit) for c in changes):
        _hdr("Batch (full recheck)", "Batch contains DomainModelBaseEdit — rechecking everything", json_only=json_only)
        check_domain_model(spec, json_only=json_only)
        check_predicates(spec, json_only=json_only)
        check_transitions(spec, json_only=json_only)
        check_scenarios(spec, json_only=json_only)
        check_traces(spec, art_cont=art_cont, art_map=art_map, traces=[], json_only=json_only)
        _try_complement_and_reachability(spec, json_only=json_only)
        _note("[green]✓[/green] Batch analysis complete (full recheck)", json_only=json_only)
        return

    # Collect affected names/ids across all changes
    pred_names: list[str] = []
    trans_names: list[str] = []
    scenario_names: list[str] = []
    trace_ids: list[str] = []
    full_scenario_refresh = False
    full_predicate_refresh = False

    for c in changes:
        if isinstance(c, (PredicateAdd, PredicateEdit, PredicateRemove)):
            pred_names.append(c.pred_name)
        if isinstance(c, (TransitionAdd, TransitionEdit, TransitionRemove)):
            trans_names.append(c.trans_name)
        if isinstance(c, (ScenarioAdd, ScenarioEdit)):
            scenario_names.append(c.scenario_name)
        if isinstance(c, ScenarioRemove):
            full_scenario_refresh = True
            full_predicate_refresh = True
        if isinstance(c, (AddTestTrace, EditTestTrace)):
            if hasattr(c, "art_id") and c.art_id is not None:
                trace_ids.append(c.art_id)
        if isinstance(c, (AddLogTrace, EditLogTrace)):
            if hasattr(c, "art_id") and c.art_id is not None:
                trace_ids.append(c.art_id)

    has_logic_changes = pred_names or trans_names or scenario_names or full_scenario_refresh

    if not has_logic_changes and not trace_ids:
        _note("[dim]Batch: no logic-relevant changes, skipping analysis[/dim]", json_only=json_only)
        return

    _hdr("Batch analysis", f"{len(changes)} changes", json_only=json_only)

    # Predicates
    if pred_names or full_predicate_refresh:
        if full_predicate_refresh:
            check_predicates(spec, json_only=json_only)
        else:
            check_predicates(spec, names=pred_names, json_only=json_only)

    # Transitions
    if trans_names:
        check_transitions(spec, names=trans_names, json_only=json_only)

    # Scenarios — affected by pred/trans/scenario changes
    if has_logic_changes:
        if full_scenario_refresh:
            check_scenarios(spec, json_only=json_only)
        else:
            check_scenarios(
                spec,
                names=scenario_names,
                predicates=pred_names,
                transitions=trans_names,
                json_only=json_only,
            )

    # Traces
    if trace_ids or has_logic_changes:
        check_traces(
            spec=spec,
            art_cont=art_cont,
            art_map=art_map,
            traces=trace_ids,
            json_only=json_only,
        )

    # Complement + reachability (once)
    if has_logic_changes:
        _try_complement_and_reachability(spec, json_only=json_only)

    _note("[green]✓[/green] Batch analysis complete", json_only=json_only)


def _try_complement_and_reachability(spec: Spec, json_only: bool = False) -> None:
    """Helper: run complement + reachability, swallowing non-fatal errors."""
    try:
        maybe_update_spec_complement(spec)
    except Exception as e:
        _note(f"[yellow]⚠[/yellow] Complement update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)

    try:
        update_reachability_graph(spec, json_only=json_only)
    except Exception as e:
        _note(f"[yellow]⚠[/yellow] Reachability update failed (non-fatal): {type(e).__name__}: {e}", json_only=json_only)


#
#   Imandra Inc.
#
#   speclogician/presentation/api.py
#

from __future__ import annotations

# --- Controls how we print things out (and also contains some data)
from speclogician.presentation.ctx import RenderCtx

# --- Underlying models
from speclogician.modeling.predicates import ActionPredicate, StatePredicate, Transition
from speclogician.modeling.domain import DomainModel
from speclogician.modeling.spec import Spec
from speclogician.modeling.complement import ScenarioComplement
from speclogician.data.traces import TestTrace, LogTrace
from speclogician.data.refs import DocRef, SrcCodeRef
from speclogician.data.container import ArtifactContainer, SmartFindResults
from speclogician.state.inst import StateInstance
from speclogician.state.diff import StateDiff
from speclogician.modeling.scenario import Scenario
from speclogician.state.recommender import Recommendation
from speclogician.modeling.trace_reachability import TraceReachabilityAnalysis

# --- Builders
from speclogician.presentation.builders.predicate import (
    build_predicate_pm,
    build_transition_pm,
)
from speclogician.presentation.builders.domain import build_domain_model_pm
from speclogician.presentation.builders.trace import build_test_trace_pm, build_log_trace_pm
from speclogician.presentation.builders.spec import build_spec_pm
from speclogician.presentation.builders.scenario import build_scenario_pm
from speclogician.presentation.builders.data_artifact import build_doc_ref_pm, build_src_code_ref_pm
from speclogician.presentation.builders.container import build_artifact_container_pm
from speclogician.presentation.builders.smart_find import build_smart_find_pm
from speclogician.presentation.builders.state_instance import build_state_instance_pm
from speclogician.presentation.builders.state_diff import build_state_diff_pm
from speclogician.presentation.builders.instances_list import build_instances_list_pm
from speclogician.presentation.builders.recommendations import build_recommendations_pm
from speclogician.presentation.builders.scenario_complement import build_scenario_complement_pm
from speclogician.presentation.builders.trace_reachability import build_trace_reachability_pm

# --- Renderers
from speclogician.presentation.renderers.predicate import (
    render_predicate,
    render_transition,
)
from speclogician.presentation.renderers.trace import render_test_trace, render_log_trace
from speclogician.presentation.renderers.domain import render_domain_model
from speclogician.presentation.renderers.spec import render_spec
from speclogician.presentation.renderers.scenario import render_scenario
from speclogician.presentation.renderers.data_artifact import render_doc_ref, render_src_code_ref
from speclogician.presentation.renderers.container import render_artifact_container
from speclogician.presentation.renderers.smart_find import render_smart_find
from speclogician.presentation.renderers.state_instance import render_state_instance
from speclogician.presentation.renderers.state_diff import render_state_diff
from speclogician.presentation.renderers.instances_list import render_instances_list
from speclogician.presentation.renderers.recommendations import render_recommendations
from speclogician.presentation.renderers.scenario_complement import render_scenario_complement
from speclogician.presentation.renderers.trace_reachability import render_trace_reachability


def present_predicate(
    p: StatePredicate | ActionPredicate,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_predicate_pm(p, ctx)
    if json_only:
        return pm.model_dump()
    return render_predicate(pm, ctx=ctx)


def present_transition(
    t: Transition,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_transition_pm(t, ctx)
    if json_only:
        return pm.model_dump()
    return render_transition(pm, ctx=ctx)


def present_domain_model(
    dm: DomainModel,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_domain_model_pm(dm, ctx)
    if json_only:
        return pm.model_dump()
    return render_domain_model(pm, ctx=ctx)


def present_spec(
    spec: Spec,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_spec_pm(spec, ctx)
    if json_only:
        return pm.model_dump()
    return render_spec(pm, ctx=ctx)


def present_scenario_complement(
    comp: ScenarioComplement,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_scenario_complement_pm(comp, ctx)
    if json_only:
        return pm.model_dump()
    return render_scenario_complement(pm, ctx=ctx)


def present_trace(
    tr: TestTrace | LogTrace,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    if isinstance(tr, TestTrace):
        pm = build_test_trace_pm(tr, ctx)
        if json_only:
            return pm.model_dump()
        return render_test_trace(pm, ctx=ctx)

    if isinstance(tr, LogTrace):
        pm = build_log_trace_pm(tr, ctx)
        if json_only:
            return pm.model_dump()
        return render_log_trace(pm, ctx=ctx)

    raise TypeError(f"Unsupported trace type: {type(tr).__name__}")


def present_data_artifact(
    a: DocRef | SrcCodeRef,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    if isinstance(a, DocRef):
        pm = build_doc_ref_pm(a, ctx)
        if json_only:
            return pm.model_dump()
        return render_doc_ref(pm, ctx=ctx)

    if isinstance(a, SrcCodeRef):
        pm = build_src_code_ref_pm(a, ctx)
        if json_only:
            return pm.model_dump()
        return render_src_code_ref(pm, ctx=ctx)

    raise TypeError(f"Unsupported data artifact type: {type(a).__name__}")


def present_artifact_container(
    cont: ArtifactContainer,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_artifact_container_pm(cont, ctx)
    if json_only:
        return pm.model_dump()
    return render_artifact_container(pm, ctx=ctx)


def present_smart_find(
    res: SmartFindResults,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_smart_find_pm(res, ctx)
    if json_only:
        return pm.model_dump()
    return render_smart_find(pm, ctx=ctx)


def present_state_diff(
    sd: StateDiff,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_state_diff_pm(sd, ctx)
    if json_only:
        return pm.model_dump()
    return render_state_diff(pm, ctx=ctx)


def present_state_instance(
    si: StateInstance,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_state_instance_pm(si, ctx)
    if json_only:
        return pm.model_dump()
    return render_state_instance(pm, ctx=ctx)


def present_scenario(
    sc: Scenario,
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_scenario_pm(sc, ctx)
    if json_only:
        return pm.model_dump()
    return render_scenario(pm, ctx=ctx)


def present_instances_list(
    states: list[StateInstance],
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_instances_list_pm(states, ctx)
    if json_only:
        return pm.model_dump()
    return render_instances_list(pm, ctx=ctx)


def present_recommendations(
    recs: list[Recommendation],
    *,
    ctx: RenderCtx,
    json_only: bool,
):
    pm = build_recommendations_pm(recs, ctx)
    return pm.model_dump() if json_only else render_recommendations(pm, ctx=ctx)

def present_trace_reachability(
    ana: TraceReachabilityAnalysis | None,
    *,
    art_cont: ArtifactContainer,
    ctx: RenderCtx,
    json_only: bool,
):
    """
    Present the trace-reachability analysis.

    NOTE: Needs `art_cont` to derive human-friendly trace labels (test name / log filename).
    """
    pm = build_trace_reachability_pm(ana, art_cont, ctx)
    if pm is None:
        return None if json_only else None
    return pm.model_dump() if json_only else render_trace_reachability(pm, ctx=ctx)


#   
#   Imandra Inc.
#
#   speclogician/presentation/builders/predicate.py
#

from __future__ import annotations

from speclogician.modeling.predicates import ActionPredicate, PredicateType, StatePredicate, Transition
from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.builders._links import build_links_pm
from speclogician.presentation.models.predicate import PredicatePM, TransitionPM


def build_predicate_pm(p: StatePredicate | ActionPredicate, ctx: RenderCtx) -> PredicatePM:
    pred_type = PredicateType.STATE if isinstance(p, StatePredicate) else PredicateType.ACTION
    signature = f"{p.name} s" if pred_type == PredicateType.STATE else f"{p.name} s a"

    # Prefer to_iml if present (yours is)
    iml = p.to_iml() if hasattr(p, "to_iml") else ""

    return PredicatePM(
        kind=p.kind,  # "state_predicate" | "action_predicate"
        comp_id=p.comp_id,
        name=p.name,
        last_updated=p.last_updated,
        pred_type=pred_type,
        signature=signature,
        src_code=p.src_code or "",
        iml=iml,
        is_iml_valid=p.is_iml_valid,
        iml_error=p.iml_error,
        links=build_links_pm(p.comp_id, ctx),
    )


def build_transition_pm(t: Transition, ctx: RenderCtx) -> TransitionPM:
    signature = f"{t.name} s a"
    iml = t.to_iml() if hasattr(t, "to_iml") else ""

    return TransitionPM(
        comp_id=t.comp_id,
        name=t.name,
        last_updated=t.last_updated,
        signature=signature,
        src_code=t.src_code or "",
        iml=iml,
        is_iml_valid=t.is_iml_valid,
        iml_error=t.iml_error,
        links=build_links_pm(t.comp_id, ctx),
    )


#
#   Imandra Inc.
#   
#   speclogician/logic/lib/predicates.py
#

import asyncio

from rich.progress import Progress

from speclogician.modeling.spec import Spec
from speclogician.logic.api import SpecLogicianImandraX
from speclogician.utils import IMLValidity, console

def check_predicates(
    spec: Spec,
    names: list[str] = [],
    json_only: bool = False,
) -> None:
    """
    Check predicate IML validity via ImandraX and write results into predicate objects:
        - predicate.is_iml_valid
        - predicate.iml_error
    If `names` is non-empty, checks only those predicate names; otherwise checks all.
    """

    dm = spec.domain_model

    # Resolve which predicates to check (by name), across state+action predicates.
    all_state = list(dm.state_preds)
    all_action = list(dm.action_preds)

    if names:
        name_set = set(names)
        state_to_check = [p for p in all_state if p.name in name_set]
        action_to_check = [p for p in all_action if p.name in name_set]

        # If a name was requested but doesn't exist, we can't attach the result to a predicate.
        # We *log* it (non-json) but do not raise.
        if not json_only:
            known = {p.name for p in all_state} | {p.name for p in all_action}
            missing = [n for n in names if n not in known]
            if missing:
                console.print(f"[yellow]Logician[/yellow] Unknown predicate name(s) requested: {missing}")
    else:
        state_to_check = all_state
        action_to_check = all_action

    # Helper: stringify ImandraX eval_model errors (best effort)
    def _errs_to_str(res: object) -> str:
        es = getattr(res, "errors", None) or []
        if es:
            return "\n".join(str(e) for e in es)
        return "ImandraX eval_model: has_errors=True"

    async def _check_all() -> None:
        async with SpecLogicianImandraX() as ix:
            # --- State predicates ---
            for p in state_to_check:
                src = dm.make_iml_model([p.name], [], [])
                try:
                    r = await ix.load_model(src)
                    has_err = bool(getattr(r, "has_errors", False))
                    p.is_iml_valid = IMLValidity.INVALID if has_err else IMLValidity.VALID
                    p.iml_error = None if not has_err else _errs_to_str(r)
                except Exception as e:
                    p.is_iml_valid = IMLValidity.UNKNOWN
                    p.iml_error = str(e)

            # --- Action predicates ---
            for p in action_to_check:
                src = dm.make_iml_model([], [p.name], [])
                try:
                    r = await ix.load_model(src)
                    has_err = bool(getattr(r, "has_errors", False))
                    p.is_iml_valid = IMLValidity.INVALID if has_err else IMLValidity.VALID
                    p.iml_error = None if not has_err else _errs_to_str(r)
                except Exception as e:
                    p.is_iml_valid = IMLValidity.UNKNOWN
                    p.iml_error = str(e)

    if not json_only:
        console.print("[bold cyan]Logician[/bold cyan] Checking predicate IML validity")

    # Show progress only in non-json mode
    if not json_only:
        with Progress(transient=True) as progress:
            t_state = progress.add_task("Checking state predicates", total=len(state_to_check))
            t_action = progress.add_task("Checking action predicates", total=len(action_to_check))

            # We still do the actual work in one ImandraX session; progress is updated as we go.
            async def _check_with_progress() -> None:
                async with SpecLogicianImandraX() as ix:
                    for p in state_to_check:
                        src = dm.make_iml_model([p.name], [], [])
                        try:
                            r = await ix.load_model(src)
                            has_err = bool(getattr(r, "has_errors", False))
                            p.is_iml_valid = IMLValidity.INVALID if has_err else IMLValidity.VALID
                            p.iml_error = None if not has_err else _errs_to_str(r)
                        except Exception as e:
                            p.is_iml_valid = IMLValidity.UNKNOWN
                            p.iml_error = str(e)
                        progress.advance(t_state)

                    for p in action_to_check:
                        src = dm.make_iml_model([], [p.name], [])
                        try:
                            r = await ix.load_model(src)
                            has_err = bool(getattr(r, "has_errors", False))
                            p.is_iml_valid = IMLValidity.INVALID if has_err else IMLValidity.VALID
                            p.iml_error = None if not has_err else _errs_to_str(r)
                        except Exception as e:
                            p.is_iml_valid = IMLValidity.UNKNOWN
                            p.iml_error = str(e)
                        progress.advance(t_action)

            try:
                asyncio.run(_check_with_progress())
            except Exception as e:
                # If the whole checking run fails (e.g. credentials/network),
                # we don't have per-predicate errors; mark checked ones as unknown.
                for p in state_to_check + action_to_check:
                    p.is_iml_valid = IMLValidity.UNKNOWN
                    p.iml_error = str(e)
    else:
        # No progress bar, just run.
        try:
            asyncio.run(_check_all())
        except Exception as e:
            for p in state_to_check + action_to_check:
                p.is_iml_valid = IMLValidity.UNKNOWN
                p.iml_error = str(e)

    # Refresh domain-model predicate stats (global, not just filtered set)
    dm.num_state_preds_total = len(all_state)
    dm.num_action_preds_total = len(all_action)

    dm.num_state_preds_valid_logic = sum(1 for p in all_state if getattr(p, "is_iml_valid", IMLValidity.UNKNOWN) == IMLValidity.VALID)
    dm.num_action_preds_valid_logic = sum(1 for p in all_action if getattr(p, "is_iml_valid", IMLValidity.UNKNOWN) == IMLValidity.VALID)

    dm.num_preds_total = dm.num_state_preds_total + dm.num_action_preds_total
    dm.num_preds_valid_logic = dm.num_state_preds_valid_logic + dm.num_action_preds_valid_logic

    if not json_only:
        console.print(
            f"[green]✓[/green] Predicates checked. "
            f"state_valid={dm.num_state_preds_valid_logic}/{dm.num_state_preds_total}, "
            f"action_valid={dm.num_action_preds_valid_logic}/{dm.num_action_preds_total}"
        )
    



#
#   Imandra Inc.
#
#   speclogician/cli/view_cmd.py
#

from __future__ import annotations

import json
from typing import Optional
import typer

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.api import (
    present_state_instance,
    present_spec,
    present_domain_model,
    present_artifact_container,
    present_state_diff,
    present_instances_list,
    present_trace_reachability,
)

from .utils import get_state

app = typer.Typer(help="View components of the current state (rich or JSON).")


def _mk_ctx(
    *,
    compact: bool,
    show_stats_only: bool,
    show_code: bool,
    show_ids: bool,
    show_art_ids: bool,
    trace_id: str | None = None,
) -> RenderCtx:
    return RenderCtx(
        compact=compact,
        show_stats_only=show_stats_only,
        show_code=show_code,
        show_ids=show_ids,
        show_art_ids=show_art_ids,
        trace_id=trace_id,
    )

def _emit(obj: object, *, json_only: bool) -> None:
    if json_only:
        typer.echo(json.dumps(obj, indent=2, default=str))
    else:
        from speclogician.utils import console
        console.print(obj)


# ---------------------------
# view state (full)
# ---------------------------

@app.command("state")
def view_state(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json", help="Emit JSON only"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()


    rctx = _mk_ctx(
        compact=compact,
        show_stats_only=show_stats_only,
        show_code=show_code,
        show_ids=show_ids,
        show_art_ids=show_art_ids,
    )

    out = present_state_instance(si, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view trace reachability
# ---------------------------

@app.command("trace-reachability")
def view_trace_reachability_cmd(
    ctx: typer.Context,
    trace: Optional[str] = typer.Option(None, "--trace", help="Filter to a single trace id"),
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()

    if si.trace_reachability is None:
        raise typer.BadParameter("No trace_reachability available on this instance.")

    rctx = _mk_ctx(
        compact=compact,
        show_stats_only=show_stats_only,
        show_code=show_code,
        show_ids=show_ids,
        show_art_ids=show_art_ids,
        trace_id=trace,  # NEW
    )

    # you’ll add this entrypoint in presentation/api.py
    out = present_trace_reachability(si.trace_reachability, art_cont=si.art_container, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view spec
# ---------------------------

@app.command("spec")
def view_spec_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()

    rctx = _mk_ctx(compact=compact, show_stats_only=show_stats_only, show_code=show_code, show_ids=show_ids, show_art_ids=show_art_ids)
    out = present_spec(si.spec, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view domain model
# ---------------------------

@app.command("domain")
def view_domain_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()

    rctx = _mk_ctx(compact=compact, show_stats_only=show_stats_only, show_code=show_code, show_ids=show_ids, show_art_ids=show_art_ids)
    out = present_domain_model(si.spec.domain_model, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view artifacts container
# ---------------------------

@app.command("artifacts")
def view_artifacts_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()

    rctx = _mk_ctx(compact=compact, show_stats_only=show_stats_only, show_code=show_code, show_ids=show_ids, show_art_ids=show_art_ids)
    out = present_artifact_container(si.art_container, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view diff
# ---------------------------

@app.command("diff")
def view_diff_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    si = st.curr_state()

    if si.state_diff is None:
        raise typer.BadParameter("No state_diff available on this instance.")

    rctx = _mk_ctx(compact=compact, show_stats_only=show_stats_only, show_code=show_code, show_ids=show_ids, show_art_ids=show_art_ids)
    out = present_state_diff(si.state_diff, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


# ---------------------------
# view list of instances
# ---------------------------

@app.command("instances")
def view_instances_cmd(
    ctx: typer.Context,
    json_only: bool = typer.Option(False, "--json"),
    compact: bool = typer.Option(False, "--compact"),
    show_stats_only: bool = typer.Option(False, "--stats-only"),
    show_code: bool = typer.Option(True, "--code/--no-code"),
    show_ids: bool = typer.Option(False, "--ids"),
    show_art_ids: bool = typer.Option(False, "--art-ids"),
) -> None:
    st = get_state(ctx)
    rctx = _mk_ctx(compact=compact, show_stats_only=show_stats_only, show_code=show_code, show_ids=show_ids, show_art_ids=show_art_ids)
    out = present_instances_list(st.instances, ctx=rctx, json_only=json_only)
    _emit(out, json_only=json_only)


#
#   Imandra Inc.
#
#   speclogician/presentation/preview/fixtures/state_instance_summary.py
#

from __future__ import annotations

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.state_instance_summary import StateInstanceSummaryPM
from speclogician.presentation.renderers.state_instance_summary import render_state_instance_summary


def make_state_instance_summary_pm(ctx: RenderCtx, variant: str) -> StateInstanceSummaryPM:
    if variant == "minimal":
        return StateInstanceSummaryPM(
            created_at="1970-01-01T00:00:00Z",
            num_changes=0,
            base_status="UNKNOWN",
            base_has_state=False,
            base_has_action=False,
            num_preds_total=0,
            num_preds_matched=0,
            num_trans_total=0,
            num_trans_valid_logic=0,
            num_sc_total=0,
            num_sc_missing=0,
            num_sc_inconsistent=0,
            num_sc_conflicted=0,
            num_test_traces_total=0,
            num_test_traces_logic_good=0,
            num_log_traces_total=0,
            num_log_traces_logic_good=0,
        )

    if variant == "edge":
        return StateInstanceSummaryPM(
            created_at="2026-01-09T10:12:00Z",
            num_changes=1234,
            base_status="VALID",
            base_has_state=True,
            base_has_action=True,
            num_preds_total=999,
            num_preds_matched=888,
            num_trans_total=77,
            num_trans_valid_logic=66,
            num_sc_total=55,
            num_sc_missing=44,
            num_sc_inconsistent=33,
            num_sc_conflicted=22,
            num_test_traces_total=111,
            num_test_traces_logic_good=100,
            num_log_traces_total=200,
            num_log_traces_logic_good=199,
        )

    # typical
    return StateInstanceSummaryPM(
        created_at="2026-01-09T10:12:00Z",
        num_changes=2,
        base_status="VALID",
        base_has_state=True,
        base_has_action=True,
        num_preds_total=12,
        num_preds_matched=7,
        num_trans_total=4,
        num_trans_valid_logic=3,
        num_sc_total=9,
        num_sc_missing=1,
        num_sc_inconsistent=0,
        num_sc_conflicted=0,
        num_test_traces_total=8,
        num_test_traces_logic_good=6,
        num_log_traces_total=4,
        num_log_traces_logic_good=4,
    )


def render(pm: StateInstanceSummaryPM, ctx: RenderCtx):
    return render_state_instance_summary(pm, ctx=ctx)



#
#   Imandra Inc.
#
#   speclogician/data/artifact.py
#

from __future__ import annotations

import uuid
from pydantic import BaseModel, Field, field_validator

class Artifact(BaseModel):
    """
    Base class for an Artifact
    """
    # Assigned task ID, new one created if not provided
    art_id : str | None = Field(default_factory=lambda: str(uuid.uuid4()))

    @field_validator("art_id", mode="before")
    @classmethod
    def _coerce_none_to_uuid(cls, v:str|None):
        if v is None:
            return str(uuid.uuid4())
        return v

class DataArtifact(Artifact):
    """
    Data may be either source code or documentation
    """
    pass


#
#   Imandra Inc.
#
#   speclogician/presentation/builders/domain.py
#

from __future__ import annotations

from speclogician.modeling.domain import DomainModel
from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.domain import (
    DomainModelPM,
    DomainCountsPM,
    DomainCodePM,
)


def build_domain_model_pm(dm: DomainModel, ctx: RenderCtx) -> DomainModelPM:
    counts = DomainCountsPM(
        base_status=dm.base_status,
        base_has_state=dm.base_has_state,
        base_has_action=dm.base_has_action,
        num_state_preds_total=dm.num_state_preds_total,
        num_state_preds_valid_logic=dm.num_state_preds_valid_logic,
        num_state_preds_matched=dm.num_state_preds_matched,
        num_action_preds_total=dm.num_action_preds_total,
        num_action_preds_valid_logic=dm.num_action_preds_valid_logic,
        num_action_preds_matched=dm.num_action_preds_matched,
        num_preds_total=dm.num_preds_total,
        num_preds_valid_logic=dm.num_preds_valid_logic,
        num_preds_matched=dm.num_preds_matched,
        num_trans_total=dm.num_trans_total,
        num_trans_valid_logic=dm.num_trans_valid_logic,
        num_trans_matched=dm.num_trans_matched,
    )

    # Always construct code PM for stable JSON; keep fields empty unless requested.
    code = DomainCodePM()

    if not ctx.show_stats_only and ctx.show_code and not ctx.compact:
        # Base
        if getattr(ctx, "show_base", True):
            code.base = dm.base or ""

        # Predicates
        if getattr(ctx, "show_predicates", True):
            code.state_preds_iml = "\n\n".join(p.to_iml() for p in dm.state_preds).strip()
            code.action_preds_iml = "\n\n".join(p.to_iml() for p in dm.action_preds).strip()

        # Transitions
        if getattr(ctx, "show_transitions", True):
            code.transitions_iml = "\n\n".join(t.to_iml() for t in dm.transitions).strip()

    return DomainModelPM(counts=counts, code=code)


#
#   Imandra Inc.
#
#   speclogician/presentation/preview/fixtures/smart_find.py
#

from __future__ import annotations

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.smart_find import (
    SmartFindPM,
    SmartFindCountsPM,
    SmartFindItemsPM,
)

from speclogician.presentation.preview.fixtures.trace import (
    minimal_test_trace_pm,
    typical_test_trace_pm,
    edge_test_trace_pm,
    minimal_log_trace_pm,
    typical_log_trace_pm,
    edge_log_trace_pm,
)

from speclogician.presentation.preview.fixtures.data_artifact import (
    minimal_doc_ref_pm,
    typical_doc_ref_pm,
    edge_doc_ref_pm,
    minimal_src_code_ref_pm,
    typical_src_code_ref_pm,
    edge_src_code_ref_pm,
)


# -----------------------------------------------------------------------------
# variants
# -----------------------------------------------------------------------------

def _smart_find_minimal() -> SmartFindPM:
    needle = ""
    return SmartFindPM(
        counts=SmartFindCountsPM(
            needle=needle,
            num_test_traces=0,
            num_log_traces=0,
            num_doc_refs=0,
            num_src_code_refs=0,
        ),
        items=SmartFindItemsPM(
            test_traces=[],
            log_traces=[],
            doc_refs=[],
            src_code_refs=[],
        ),
    )


def _smart_find_typical() -> SmartFindPM:
    needle = "order size"
    tt = typical_test_trace_pm(with_contents=False, invalid=False)
    lt = typical_log_trace_pm(with_contents=False, invalid=False)
    dr = typical_doc_ref_pm()
    sr = typical_src_code_ref_pm()

    return SmartFindPM(
        counts=SmartFindCountsPM(
            needle=needle,
            num_test_traces=1,
            num_log_traces=1,
            num_doc_refs=1,
            num_src_code_refs=1,
        ),
        items=SmartFindItemsPM(
            test_traces=[tt],
            log_traces=[lt],
            doc_refs=[dr],
            src_code_refs=[sr],
        ),
    )


def _smart_find_edge() -> SmartFindPM:
    needle = "pred_eq_ten OR max_qty (line:42)"
    # Stress renderers: mix long contents + invalid + lots of items
    tts = [
        typical_test_trace_pm(with_contents=True, invalid=True),
        edge_test_trace_pm(),
    ]
    lts = [
        typical_log_trace_pm(with_contents=True, invalid=True),
        edge_log_trace_pm(),
    ]
    drs = [
        edge_doc_ref_pm(),
        typical_doc_ref_pm(),
        minimal_doc_ref_pm(),
    ]
    srs = [
        edge_src_code_ref_pm(),
        typical_src_code_ref_pm(),
        minimal_src_code_ref_pm(),
    ]

    return SmartFindPM(
        counts=SmartFindCountsPM(
            needle=needle,
            num_test_traces=len(tts),
            num_log_traces=len(lts),
            num_doc_refs=len(drs),
            num_src_code_refs=len(srs),
        ),
        items=SmartFindItemsPM(
            test_traces=tts,
            log_traces=lts,
            doc_refs=drs,
            src_code_refs=srs,
        ),
    )


# -----------------------------------------------------------------------------
# top-level fixture entrypoint (matches your preview setup)
# -----------------------------------------------------------------------------

def make_smart_find_pm(ctx: RenderCtx, variant: str = "typical") -> SmartFindPM:
    v = (variant or "typical").strip().lower()
    if v == "minimal":
        return _smart_find_minimal()
    if v == "edge":
        return _smart_find_edge()
    return _smart_find_typical()


# -----------------------------------------------------------------------------
# renderer adapter (matches other fixtures style)
# -----------------------------------------------------------------------------

def render(pm: SmartFindPM, ctx: RenderCtx):
    from speclogician.presentation.renderers.smart_find import render_smart_find
    return render_smart_find(pm, ctx=ctx)


#
#   Imandra Inc.
#
#   speclogician/state/change_result.py
#

from __future__ import annotations

from typing import Literal, TypedDict, Union

from ..state.change import StateChange
from ..state.inst import StateInstance
from ..state.diff import StateDiff

class ProcessChangeError(TypedDict):
    ok: Literal[False]
    stage: Literal["apply_change", "analyze_change", "calc_diff", "refresh_stats", "trace_reachability"]
    error: Exception
    change: StateChange
    old_state: StateInstance

class ProcessChangeSuccess(TypedDict):
    ok: Literal[True]
    stage: Literal["ok"]
    change: StateChange
    old_state: StateInstance
    new_state: StateInstance
    state_diff: StateDiff | None # this is because it could potentially be the first state instance
    new_num_instances: int

ProcessChangeResult = Union[ProcessChangeError, ProcessChangeSuccess]


# ---------------------------------------------------------------------------
# Batch results
# ---------------------------------------------------------------------------

class ProcessBatchError(TypedDict):
    ok: Literal[False]
    stage: Literal["apply_change", "analyze_change", "calc_diff", "refresh_stats", "trace_reachability"]
    error: Exception
    changes: list[StateChange]
    change_index: int          # index of the failing change (-1 for post-apply stages)
    old_state: StateInstance

class ProcessBatchSuccess(TypedDict):
    ok: Literal[True]
    stage: Literal["ok"]
    changes: list[StateChange]
    old_state: StateInstance
    new_state: StateInstance
    state_diff: StateDiff | None
    new_num_instances: int

ProcessBatchResult = Union[ProcessBatchError, ProcessBatchSuccess]


#
#   Imandra Inc.
#
#   speclogician/presentation/models/smart_find.py
#

from __future__ import annotations

from typing import Literal
from pydantic import BaseModel, Field

from speclogician.presentation.models.trace import TestTracePM, LogTracePM
from speclogician.presentation.models.data_artifact import DocRefPM, SrcCodeRefPM


class SmartFindCountsPM(BaseModel):
    needle: str
    num_test_traces: int = 0
    num_log_traces: int = 0
    num_doc_refs: int = 0
    num_src_code_refs: int = 0


class SmartFindItemsPM(BaseModel):
    test_traces: list[TestTracePM] = Field(default_factory=list)
    log_traces: list[LogTracePM] = Field(default_factory=list)
    doc_refs: list[DocRefPM] = Field(default_factory=list)
    src_code_refs: list[SrcCodeRefPM] = Field(default_factory=list)


class SmartFindPM(BaseModel):
    kind: Literal["smart_find"] = "smart_find"
    counts: SmartFindCountsPM
    items: SmartFindItemsPM


#
#   Imandra Inc.
#
#   speclogician/presentation/renderers/spec.py
#

from __future__ import annotations

from rich.console import Group
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from rich import box

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.spec import SpecPM


def _scenario_stats_table(pm: SpecPM, *, ctx: RenderCtx) -> Table:
    b = box.MINIMAL if ctx.compact else box.SIMPLE
    t = Table(show_header=False, box=b, show_edge=False, pad_edge=False)
    t.add_column(style="italic dim", no_wrap=True)
    t.add_column(justify="right", no_wrap=True)

    c = pm.counts

    t.add_row(Text("Scenarios", style="bold magenta"), Text(""))
    t.add_row("Total", Text(str(c.num_sc_total), style="bold"))
    t.add_row("Missing components", Text(str(c.num_sc_missing)))
    t.add_row("Matched", Text(str(c.num_sc_matched)))
    t.add_row("Inconsistent", Text(str(c.num_sc_inconsistent)))
    t.add_row("Conflicted", Text(str(c.num_sc_conflicted)))

    return t


def _scenario_list_table(pm: SpecPM, *, ctx: RenderCtx) -> Table:
    b = box.MINIMAL if ctx.compact else box.SIMPLE
    t = Table(title="Scenario list", box=b, show_edge=False, pad_edge=False, expand=False)
    t.add_column("Name", no_wrap=True)

    if not pm.scenario_names:
        t.add_row(Text("(none)", style="dim"))
        return t

    for name in pm.scenario_names:
        t.add_row(Text(name))

    return t


def _scenario_complement_table(pm: SpecPM, *, ctx: RenderCtx) -> Table:
    """
    Lightweight complement summary (counts + optional preview list).
    Full regions are intentionally not rendered here (they can be huge).
    """
    b = box.MINIMAL if ctx.compact else box.SIMPLE
    t = Table(show_header=False, box=b, show_edge=False, pad_edge=False)
    t.add_column(style="italic dim", no_wrap=True)
    t.add_column(justify="right", no_wrap=True)

    sc = getattr(pm, "scenario_complement", None)
    if sc is None:
        t.add_row(Text("Complement", style="bold magenta"), Text(""))
        t.add_row(Text("(none)", style="dim"), Text(""))
        return t

    t.add_row(Text("Complement", style="bold magenta"), Text(""))
    t.add_row("Regions", Text(str(getattr(sc, "num_regions", 0)), style="bold"))

    # Optional “preview” identifiers if you included them in the PM
    preview = list(getattr(sc, "region_fps_preview", []) or [])
    if preview:
        t.add_row(Text("Preview", style="dim"), Text(""))
        for fp in preview[:3]:
            t.add_row("•", Text(str(fp), style="dim"))

    return t


def render_spec(pm: SpecPM, *, ctx: RenderCtx):
    # Domain model block (reuse your existing renderer, but DON'T wrap it in extra panels here)
    from speclogician.presentation.renderers.domain import render_domain_model

    blocks: list[object] = []

    # ---- Domain ----
    blocks.append(Text("Domain", style="bold magenta"))
    blocks.append(render_domain_model(pm.domain, ctx=ctx))

    # ---- Scenarios ----
    blocks.append(Rule(style="dim"))
    blocks.append(Text("Scenarios", style="bold magenta"))
    blocks.append(_scenario_stats_table(pm, ctx=ctx))
    blocks.append(Text(""))  # small spacer
    blocks.append(_scenario_list_table(pm, ctx=ctx))

    # ---- Complement ----
    sc = getattr(pm, "scenario_complement", None)
    if sc is not None or ctx.show_empty_sections:
        blocks.append(Rule(style="dim"))
        blocks.append(_scenario_complement_table(pm, ctx=ctx))

    return Panel(Group(*blocks), title="[bold]Spec[/bold]", border_style="magenta")


#
#   Imandra Inc.
#
#   speclogician/logic/lib/domain_model.py
#

import asyncio

from speclogician.modeling.domain import BaseStatus
from speclogician.modeling.spec import Spec
from speclogician.logic.api import SpecLogicianImandraX
from speclogician.utils import console

def check_domain_model(spec: Spec, json_only: bool = False) -> None:
    """
    Updates:
    - spec.domain_model.base_has_state
    - spec.domain_model.base_has_action
    - spec.domain_model.base_status  (via ImandraX)
    """
    dm = spec.domain_model

    if not json_only:
        console.print("[bold cyan]Logician[/bold cyan] Checking domain model base")

    # Local checks
    dm.base_has_state = dm.state_specified()
    dm.base_has_action = dm.action_specified()

    if not json_only:
        console.print(
            f"  • has state type : {'✓' if dm.base_has_state else '✗'}\n"
            f"  • has action type: {'✓' if dm.base_has_action else '✗'}"
        )

    # Build "base-only" model and ask ImandraX to load/typecheck it
    base_src = dm.make_iml_model([], [], [])

    async def _check() -> bool:
        async with SpecLogicianImandraX() as ix:
            res = await ix.load_model(base_src)
            return not bool(getattr(res, "has_errors", False))

    try:
        ok = asyncio.run(_check())
    except Exception as e:
        dm.base_status = BaseStatus.UNKNOWN
        if not json_only:
            console.print("[yellow]⚠ Unable to check base model via ImandraX[/yellow]")
            console.print(f"[dim]{e}[/dim]")
        return

    dm.base_status = BaseStatus.VALID if ok else BaseStatus.INVALID_IML

    if not json_only:
        if dm.base_status is BaseStatus.VALID:
            console.print("[green]✓ Base domain model is valid IML[/green]")
        else:
            console.print("[red]✗ Base domain model is invalid IML[/red]")



#
#   Imandra Inc.
#
#   speclogician/presentation/preview/fixtures/domain_model.py
#

from __future__ import annotations

from speclogician.modeling.domain import BaseStatus
from speclogician.presentation.models.domain import (
    DomainCountsPM,
    DomainCodePM,
    DomainModelPM,
)
from speclogician.presentation.ctx import RenderCtx


# -----------------------------------------------------------------------------
# PM builders (variants)
# -----------------------------------------------------------------------------

def make_domain_model_pm(ctx: RenderCtx, variant: str = "typical") -> DomainModelPM:
    v = (variant or "typical").lower()
    if v == "minimal":
        return minimal_domain_model_pm()
    if v == "edge":
        return edge_domain_model_pm()
    return typical_domain_model_pm()


def minimal_domain_model_pm() -> DomainModelPM:
    # Minimal = almost empty, base unknown, no code.
    counts = DomainCountsPM(
        base_status=BaseStatus.UNKNOWN,
        base_has_state=False,
        base_has_action=False,
        num_state_preds_total=0,
        num_state_preds_valid_logic=0,
        num_state_preds_matched=0,
        num_action_preds_total=0,
        num_action_preds_valid_logic=0,
        num_action_preds_matched=0,
        num_preds_total=0,
        num_preds_valid_logic=0,
        num_preds_matched=0,
        num_trans_total=0,
        num_trans_valid_logic=0,
        num_trans_matched=0,
    )
    code = DomainCodePM(
        base="",
        state_preds_iml="",
        action_preds_iml="",
        transitions_iml="",
    )
    return DomainModelPM(counts=counts, code=code)


def typical_domain_model_pm() -> DomainModelPM:
    # Typical = sane + readable, some matched/valid numbers.
    counts = DomainCountsPM(
        base_status=BaseStatus.VALID,
        base_has_state=True,
        base_has_action=True,
        # state preds
        num_state_preds_total=4,
        num_state_preds_valid_logic=3,
        num_state_preds_matched=2,
        # action preds
        num_action_preds_total=3,
        num_action_preds_valid_logic=3,
        num_action_preds_matched=1,
        # totals
        num_preds_total=7,
        num_preds_valid_logic=6,
        num_preds_matched=3,
        # transitions
        num_trans_total=2,
        num_trans_valid_logic=2,
        num_trans_matched=1,
    )
    code = DomainCodePM(
        base=(
            "type state = { x : int }\n"
            "type action = int\n"
        ),
        state_preds_iml=(
            "let pred_eq_ten (s: state) = s.x = 10\n"
            "let pred_lt_ten (s: state) = s.x < 10\n"
            "let pred_gt_zero (s: state) = s.x > 0\n"
        ),
        action_preds_iml=(
            "let act_gt_x (s: state) (a: action) = a > s.x\n"
            "let act_eq_zero (_s: state) (a: action) = a = 0\n"
        ),
        transitions_iml=(
            "let step1 (s: state) (a: action) : state = { x = s.x + a }\n"
            "let step2 (_s: state) (a: action) : state = { x = a }\n"
        ),
    )
    return DomainModelPM(counts=counts, code=code)


def edge_domain_model_pm() -> DomainModelPM:
    # Edge = stress the renderer: long code, weird counts, invalid base.
    counts = DomainCountsPM(
        base_status=BaseStatus.INVALID_IML,
        base_has_state=True,
        base_has_action=False,
        # state preds
        num_state_preds_total=12,
        num_state_preds_valid_logic=7,
        num_state_preds_matched=0,
        # action preds
        num_action_preds_total=0,
        num_action_preds_valid_logic=0,
        num_action_preds_matched=0,
        # totals
        num_preds_total=12,
        num_preds_valid_logic=7,
        num_preds_matched=0,
        # transitions
        num_trans_total=5,
        num_trans_valid_logic=2,
        num_trans_matched=0,
    )
    code = DomainCodePM(
        base=(
            "type state = {\n"
            "  x : int;\n"
            "  (* missing action type on purpose to simulate INVALID base *)\n"
            "}\n"
        ),
        state_preds_iml=(
            "(* many predicates, including a long one to wrap *)\n"
            + "\n".join(
                f"let pred_{i} (s: state) = s.x = {i}"
                for i in range(1, 10)
            )
            + "\n"
            "let pred_very_long (s: state) = "
            "(s.x > 0 && s.x < 1000000 && (s.x mod 3 = 0 || s.x mod 5 = 0))\n"
        ),
        action_preds_iml="",  # missing because base_has_action=False
        transitions_iml=(
            "(* some transitions, one intentionally sketchy *)\n"
            "let step_ok (s: state) (_a: int) : state = { x = s.x }\n"
            "let step_weird (s: state) (a: int) : state = { x = s.x + (a + 1) }\n"
            "(* malformed-ish snippet *)\n"
            "let step_bad (s: state) (a: int) : state = { x = s.x + }\n"
        ),
    )
    return DomainModelPM(counts=counts, code=code)


# -----------------------------------------------------------------------------
# renderer adapter (matches other fixtures style)
# -----------------------------------------------------------------------------

def render(pm: DomainModelPM, ctx: RenderCtx):
    from speclogician.presentation.renderers.domain import render_domain_model
    return render_domain_model(pm, ctx=ctx)


#
#   Imandra Inc.
#
#   speclogician/logic/api/traces.py
#

from __future__ import annotations

import asyncio

from .client import SpecLogicianImandraX
from speclogician.modeling.domain import DomainModel
from speclogician.modeling.scenario import Scenario
from speclogician.data.traces import TraceArtifact, TraceIMLValidity, TraceScenarioMatchResult


async def check_trace_iml_validity_async(
    tr: TraceArtifact,
    model: DomainModel,
) -> TraceIMLValidity:
    """
    Check that trace snippets typecheck against the domain model.
    """
    async with SpecLogicianImandraX() as ix:
        v = await ix.check_trace_iml_validity(tr=tr, model=model)
        tr.iml_validity = v
        return v

def check_trace_iml_validity (
    tr: TraceArtifact,
    model: DomainModel,
) -> TraceIMLValidity:
    """
    Sync wrapper around check_trace_iml_validity_async.
    """
    return asyncio.run(check_trace_iml_validity_async(tr, model))


async def check_trace_match_async(
    tr: TraceArtifact,
    s: Scenario,
    model: DomainModel,
) -> TraceScenarioMatchResult:
    """
    Check whether the trace matches a scenario (given/when predicates + transition result).
    Assumes check_trace_iml_validity is either called inside or is handled by the async method.
    """
    async with SpecLogicianImandraX() as ix:
        return await ix.check_trace_match(tr=tr, s=s, model=model)

def check_trace_match(
    tr: TraceArtifact,
    s: Scenario,
    model: DomainModel,
) -> TraceScenarioMatchResult:
    """
    Sync wrapper around check_trace_match_async.
    """
    return asyncio.run(check_trace_match_async(tr, s, model))


#
#   Imandra Inc.    
#
#   speclogician/modeling/conflict.py
#

from pydantic import BaseModel
from typing import Literal, TypeAlias

class ScenarioOverlap(BaseModel):
    """ Two scenarios overlap in constraints """
    kind: Literal["overlap"] = "overlap"
    scenario_name1 : str
    scenario_name2 : str

    details : str

class ScenarioConsumed(BaseModel):
    """ One scenario fully absorbs another """
    kind: Literal["consumed"] = "consumed"
    scenario_name1 : str
    scenario_name2 : str

    details : str

ScenarioConflict : TypeAlias = ScenarioOverlap | ScenarioConsumed



#
#   Imandra Inc.
#
#   speclogician/commands/find_cmd.py
#

from __future__ import annotations

from typing import Any

import typer
from rich.table import Table

from .utils import get_state
from ..utils import console

app = typer.Typer(help="Find various components of the state")


# -----------------------------------------------------------------------------
# helpers
# -----------------------------------------------------------------------------

def _matches_component(q: str, *, name: str, src_code: str) -> bool:
    qq = (q or "").lower()
    return (qq in (name or "").lower()) or (qq in (src_code or "").lower())


def _emit_json(payload: dict[str, Any]) -> None:
    # Keep stdout clean for machine mode (matches your convention elsewhere)
    import json as _json
    typer.echo(_json.dumps(payload, indent=2, sort_keys=True))


def _preview(src: str, n: int = 160) -> str:
    return (src or "")[:n].replace("\n", " ")


# -----------------------------------------------------------------------------
# predicates
# -----------------------------------------------------------------------------

@app.command(name="predicates", help="Search state/action predicates by name or source text")
def search_predicates(
    ctx: typer.Context,
    q: str = typer.Argument(..., help="Search string (name or source snippet)"),
    kind: str = typer.Option(
        "all",
        "--kind",
        help="Search scope: all|state|action",
    ),
    json_only: bool = typer.Option(
        False,
        "--json",
        help="Emit JSON only (no rich output)",
    ),
) -> None:
    state = get_state(ctx=ctx)
    dm = state.curr_state().spec.domain_model

    k = (kind or "all").lower().strip()
    if k not in ("all", "state", "action"):
        raise typer.BadParameter("kind must be one of: all|state|action")

    matches: list[dict[str, Any]] = []

    if k in ("all", "state"):
        for p in getattr(dm, "state_preds", []):
            name = getattr(p, "name", "")
            src_code = getattr(p, "src_code", "")
            if _matches_component(q, name=name, src_code=src_code):
                matches.append(
                    {
                        "type": "StatePredicate",
                        "kind": getattr(p, "kind", "state_predicate"),
                        "name": name,
                        "src_preview": _preview(src_code),
                    }
                )

    if k in ("all", "action"):
        for p in getattr(dm, "action_preds", []):
            name = getattr(p, "name", "")
            src_code = getattr(p, "src_code", "")
            if _matches_component(q, name=name, src_code=src_code):
                matches.append(
                    {
                        "type": "ActionPredicate",
                        "kind": getattr(p, "kind", "action_predicate"),
                        "name": name,
                        "src_preview": _preview(src_code),
                    }
                )

    # Deterministic output (nice for tests)
    matches.sort(key=lambda m: (m.get("type", ""), m.get("name", "")))

    if json_only:
        _emit_json(
            {
                "query": q,
                "kind": k,
                "count": len(matches),
                "results": matches,
            }
        )
        return

    t = Table(title=f"Predicate search for {q!r}", show_header=True, header_style="bold")
    t.add_column("Type", no_wrap=True)
    t.add_column("Name", no_wrap=True)
    t.add_column("Preview")
    for m in matches:
        prev = m["src_preview"]
        t.add_row(m["type"], m["name"], prev + ("…" if len(prev) == 160 else ""))
    console.print(t)


# -----------------------------------------------------------------------------
# transitions
# -----------------------------------------------------------------------------

@app.command(name="transitions", help="Search transition functions by name or source text")
def search_transitions(
    ctx: typer.Context,
    q: str = typer.Argument(..., help="Search string (name or source snippet)"),
    json_only: bool = typer.Option(
        False,
        "--json",
        help="Emit JSON only (no rich output)",
    ),
) -> None:
    state = get_state(ctx=ctx)
    dm = state.curr_state().spec.domain_model

    matches: list[dict[str, Any]] = []

    for tr in getattr(dm, "transitions", []):
        name = getattr(tr, "name", "")
        src_code = getattr(tr, "src_code", "")
        if _matches_component(q, name=name, src_code=src_code):
            matches.append(
                {
                    "type": "Transition",
                    "kind": getattr(tr, "kind", "transition"),
                    "name": name,
                    "src_preview": _preview(src_code),
                }
            )

    matches.sort(key=lambda m: m.get("name", ""))

    if json_only:
        _emit_json(
            {
                "query": q,
                "count": len(matches),
                "results": matches,
            }
        )
        return

    t = Table(title=f"Transition search for {q!r}", show_header=True, header_style="bold")
    t.add_column("Name", no_wrap=True)
    t.add_column("Preview")
    for m in matches:
        prev = m["src_preview"]
        t.add_row(m["name"], prev + ("…" if len(prev) == 160 else ""))
    console.print(t)


# -----------------------------------------------------------------------------
# scenarios
# -----------------------------------------------------------------------------

def _matches_scenario(q: str, scenario: Any) -> bool:
    qq = (q or "").lower()
    name = getattr(scenario, "name", "")
    if qq in (name or "").lower():
        return True
    for section in ("given", "when", "then"):
        for comp in getattr(scenario, section, []):
            if qq in (comp or "").lower():
                return True
    return False


@app.command(name="scenarios", help="Search scenarios by name or component membership (predicate/transition names)")
def search_scenarios(
    ctx: typer.Context,
    q: str = typer.Argument(..., help="Search string (scenario name, predicate name, or transition name)"),
    json_only: bool = typer.Option(
        False,
        "--json",
        help="Emit JSON only (no rich output)",
    ),
) -> None:
    state = get_state(ctx=ctx)
    scenarios = getattr(state.curr_state().spec, "scenarios", [])

    matches: list[dict[str, Any]] = []

    for sc in scenarios:
        if _matches_scenario(q, sc):
            matches.append(
                {
                    "name": getattr(sc, "name", ""),
                    "is_initial": getattr(sc, "is_initial", False),
                    "given": list(getattr(sc, "given", [])),
                    "when": list(getattr(sc, "when", [])),
                    "then": list(getattr(sc, "then", [])),
                }
            )

    matches.sort(key=lambda m: m.get("name", ""))

    if json_only:
        _emit_json(
            {
                "query": q,
                "count": len(matches),
                "results": matches,
            }
        )
        return

    t = Table(title=f"Scenario search for {q!r}", show_header=True, header_style="bold")
    t.add_column("Name", no_wrap=True)
    t.add_column("Initial", no_wrap=True)
    t.add_column("Given")
    t.add_column("When")
    t.add_column("Then")
    for m in matches:
        t.add_row(
            m["name"],
            "yes" if m["is_initial"] else "",
            ", ".join(m["given"]),
            ", ".join(m["when"]),
            ", ".join(m["then"]),
        )
    console.print(t)


# -----------------------------------------------------------------------------
# usages (reverse lookup)
# -----------------------------------------------------------------------------

@app.command(name="usages", help="Find which scenarios and artifacts reference a given component (predicate or transition)")
def search_usages(
    ctx: typer.Context,
    component: str = typer.Argument(..., help="Exact component name (predicate or transition)"),
    json_only: bool = typer.Option(
        False,
        "--json",
        help="Emit JSON only (no rich output)",
    ),
) -> None:
    state = get_state(ctx=ctx)
    inst = state.curr_state()
    spec = inst.spec
    dm = spec.domain_model
    art_map = inst.art_map

    # Determine what kind of component this is
    comp_type: str | None = None
    for p in getattr(dm, "state_preds", []):
        if getattr(p, "name", "") == component:
            comp_type = "state_predicate"
            break
    if comp_type is None:
        for p in getattr(dm, "action_preds", []):
            if getattr(p, "name", "") == component:
                comp_type = "action_predicate"
                break
    if comp_type is None:
        for tr in getattr(dm, "transitions", []):
            if getattr(tr, "name", "") == component:
                comp_type = "transition"
                break

    # Find scenarios that use this component
    scenario_usages: list[dict[str, Any]] = []
    for sc in getattr(spec, "scenarios", []):
        sections: list[str] = []
        if component in list(getattr(sc, "given", [])):
            sections.append("given")
        if component in list(getattr(sc, "when", [])):
            sections.append("when")
        if component in list(getattr(sc, "then", [])):
            sections.append("then")
        if sections:
            scenario_usages.append({
                "scenario": getattr(sc, "name", ""),
                "sections": sections,
            })
    scenario_usages.sort(key=lambda m: m["scenario"])

    # Find linked artifacts
    linked_art_ids = sorted(art_map.get_arts_for_component(component))

    if json_only:
        _emit_json(
            {
                "component": component,
                "component_type": comp_type,
                "scenarios": scenario_usages,
                "scenario_count": len(scenario_usages),
                "linked_artifact_ids": linked_art_ids,
                "linked_artifact_count": len(linked_art_ids),
            }
        )
        return

    # Rich output
    if comp_type:
        console.print(f"[bold]{component}[/bold] ({comp_type})\n")
    else:
        console.print(f"[bold]{component}[/bold] (not found in domain model)\n")

    if scenario_usages:
        t = Table(title="Scenarios using this component", show_header=True, header_style="bold")
        t.add_column("Scenario", no_wrap=True)
        t.add_column("Sections")
        for u in scenario_usages:
            t.add_row(u["scenario"], ", ".join(u["sections"]))
        console.print(t)
    else:
        console.print("[dim]No scenarios reference this component.[/dim]")

    if linked_art_ids:
        console.print(f"\n[bold]Linked artifacts:[/bold] {', '.join(linked_art_ids)}")
    else:
        console.print("\n[dim]No artifacts linked to this component.[/dim]")


# -----------------------------------------------------------------------------
# artifacts (smart_find + PM renderer)
# -----------------------------------------------------------------------------

@app.command("artifacts")
def search_artifacts(
    ctx: typer.Context,
    q: str = typer.Argument(..., help="Search string"),
    kind: str = typer.Option(
        "all",
        "--kind",
        help="Search scope: all|tests|logs|docs|src",
    ),
    json_only: bool = typer.Option(
        False,
        "--json",
        help="Emit JSON only (no rich output)",
    ),
    ids: bool = typer.Option(
        False,
        "--ids",
        help="Show artifact IDs in rich output",
    ),
    compact: bool = typer.Option(
        False,
        "--compact",
        help="Use compact tables/panels in rich output",
    ),
    show_empty: bool = typer.Option(
        False,
        "--show-empty",
        help="Render empty sections in rich output",
    ),
    max_per_type: int = typer.Option(
        50,
        "--max-per-type",
        min=1,
        help="Maximum results to show per artifact type",
    ),
) -> None:
    """Smart text search across artifacts (tests/logs/docs/src)."""
    state = get_state(ctx=ctx)
    ac = state.curr_state().art_container

    k = (kind or "all").lower().strip()
    include: set[str] | None
    if k in ("all", "*"):
        include = None
    elif k in ("tests", "test"):
        include = {"test"}
    elif k in ("logs", "log"):
        include = {"log"}
    elif k in ("docs", "doc"):
        include = {"doc"}
    elif k in ("src", "source"):
        include = {"src"}
    else:
        raise typer.BadParameter("kind must be one of: all|tests|logs|docs|src")

    # --- search ---
    res = ac.smart_find(q, include=include, max_per_type=max_per_type)

    # --- build PM + render ---
    from speclogician.presentation.ctx import RenderCtx
    from speclogician.presentation.builders.smart_find import build_smart_find_pm
    from speclogician.presentation.renderers.smart_find import render_smart_find

    rctx = RenderCtx(
        show_art_ids=ids,
        show_ids=ids,
        compact=compact,
        show_empty_sections=show_empty,  # now a real field on RenderCtx
    )

    pm = build_smart_find_pm(res, rctx)

    if json_only:
        payload = pm.model_dump()
        # test-friendly extra metadata
        payload["query"] = q
        payload["kind"] = k
        _emit_json(payload)
        return

    console.print(render_smart_find(pm, ctx=rctx))


#
#   Imandra Inc.
#
#   speclogician/presentation/builders/instances_list.py
#

from __future__ import annotations

from speclogician.presentation.ctx import RenderCtx
from speclogician.presentation.models.instances_list import InstancesListPM, InstancesListRowPM
from speclogician.presentation.builders.state_instance_summary import build_state_instance_summary_pm
from speclogician.state.inst import StateInstance


def _changes_str(si: StateInstance) -> str:
    if not (si.changes or []):
        return "-"

    parts: list[str] = []
    for ch in si.changes:
        short = getattr(ch, "short_str", None)
        parts.append(short() if callable(short) else ch.__class__.__name__)
    return ", ".join(parts)


def build_instances_list_pm(states: list[StateInstance], ctx: RenderCtx) -> InstancesListPM:
    rows: list[InstancesListRowPM] = []
    for idx, si in enumerate(states):
        state_label = "latest" if idx == 0 else str(idx)
        rows.append(
            InstancesListRowPM(
                state_label=state_label,
                changes=_changes_str(si),
                summary=build_state_instance_summary_pm(si, ctx=ctx),
            )
        )

    return InstancesListPM(rows=rows)